unit EOFileIntf;

{*******************************************************}
{                                                       }
{       EOFile Extension Interfaces                     }
{                                                       }
{       2002..2006, 2008 YT                             }
{                                                       }
{*******************************************************}

interface

{x$DEFINE EXTENSION}

uses
	Windows, ActiveX, ShlObj {$IFNDEF EXTENSION} , LObjects {$ENDIF} ;

{$IFDEF EXTENSION}
{$INCLUDE LInterfaces.inc}
{$ENDIF}

const
	InterfaceVersion = 250;

type
	PPath = ^TPath;
	TPath = record
		ItemIDList: PItemIDList;
		Directory: WideString;
		Archive: WideString;
		Extension: Pointer;
	end;

const
	EmptyPath: TPath = (ItemIDList: nil; Directory: ''; Archive: ''; Extension: nil);

type
	TFileOperation = (
		foDelete,
		foMove,
		foRename,
		foCopy,
		foLink,
		foEnter,
		foProperty);

	TFileOperatable = set of TFileOperation;

const
	AllFileOperatable = [foDelete..foProperty];

type
	IArchiver = interface;
	IExtension = interface;

	TWideCharFileName = array[0..MAX_PATH - 1] of WideChar;

	TFileKind = (fkVirtual, fkFile, fkDrive);

	PFile = ^TFile;
	TFile = record
		NameLength: Cardinal;
		Name: TWideCharFileName;
		DisplayNameLength: Cardinal;
		DisplayName: TWideCharFileName;
		Size: Int64;
		Data: Pointer;
		Number: Word;
		Operatable: TFileOperatable;
		Work: Byte;
	case Kind: TFileKind of
	fkDrive: {4 + 4 + 24 + 4 + 24 = 60}
	(
		DriveType: DWORD;
		VolumeLabelLength: Cardinal;
		VolumeLabel: array[0..11] of WideChar;
		FileSystemLength: Cardinal;
		FileSystem: array[0..11] of WideChar;
	);
	fkFile, fkVirtual: {8 + 8 + 8 + 28 + 4 + 4 = 60}
	(
		CreationTime: TFileTime;
		LastWriteTime: TFileTime;
		LastAccessTime: TFileTime;
		ShortNameLength: Cardinal;
		ShortName: array[0..13] of WideChar;
		Attributes: DWORD;
	);
	end;

	TFolderPossibilities = set of (fpSort, fpPlaceFile, fpHasParent);

	IFolder = interface(IObservable)
		['{065E6DD3-6977-4A1B-8BEF-9A916F9A3D30}']
		function GetPath: TPath;
		function GetItems(AIndex: Integer): PFile;
		function GetCount: Integer;
		function GetValid: Boolean;
		function GetDisplayText: WideString;
		function GetPossibilities: TFolderPossibilities;
		procedure Refresh;
		procedure Move(const NewPath: WideString);
		procedure StopSurveillance;
		procedure ResumeSurveillance(Wait: Boolean);
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		function ConcatenatePath(AFile: PFile): TPath;
		function GetInfoText(AFile: PFile): WideString;
		property Path: TPath read GetPath;
		property Items[AIndex: Integer]: PFile read GetItems;
		property Count: Integer read GetCount;
		property Valid: Boolean read GetValid;
		property DisplayText: WideString read GetDisplayText;
		property Possibilities: TFolderPossibilities read GetPossibilities;
	end;

	IDirectory = interface(IFolder)
		['{8D98B7A2-ED6C-4E23-9E8F-19F949CD8475}']
		function GetAtRoot: Boolean;
		property AtRoot: Boolean read GetAtRoot;
	end;

	IArchive = interface(IFolder)
		['{30FE4515-F4E0-4584-807C-FDEB2145483D}']
		procedure UnpackMemory(AFile: PFile; var Data: string);
		procedure Unpack(AFile: PFile; const DestDir: WideString);
	end;

	INameSpace = interface(IFolder)
		['{14831640-75E3-4D4C-AE12-27E595FE5E7D}']
		function AtDesktop: Boolean;
		function AtDrives: Boolean;
	end;

	IFounds = interface(IFolder)
		['{E26F487A-62DA-4ED7-8E35-5AD417148F4C}']
		function GetWildCard: WideString;
		property WildCard: WideString read GetWildCard;
	end;

	IFolderManager = interface(IInterface)
		['{0A7B2795-3552-4A4E-9541-4830DF00826B}']
		function GetFolder(const APath: TPath): IFolder;
		function GetOpenedFolder(const APath: TPath): IFolder;
		procedure Refresh(const APath: TPath; Sub: Boolean);
		procedure ChangePath(const APath: TPath; const NewName: WideString);
		function FindArchiver(const Ext: WideString): IArchiver;
		function LinkToPath(const LinkFile: WideString): TPath;
		function ToPath(const Path: WideString): TPath;
		function ToString(const Path: TPath): WideString;
	end;

	IView = interface(IInterface)
		['{19D7D2FC-18A8-431C-A622-F91B8B9432F0}']
		function GetModel: IFolder;
		procedure SetModel(const Value: IFolder);
		function GetItems(AIndex: Integer): PFile;
		function GetCount: Integer;
		function GetSelected(AIndex: Integer): Boolean;
		function GetSelectedCount: Integer;
		procedure FocusTo(const AName: WideString);
		function FocusedFile: PFile;
		procedure SetFocus;
		procedure Close;
		property Model: IFolder read GetModel write SetModel;
		property Items[AIndex: Integer]: PFile read GetItems;
		property Count: Integer read GetCount;
		property Selected[AIndex: Integer]: Boolean read GetSelected;
		property SelectedCount: Integer read GetSelectedCount;
	end;

	CEachView = function(const View: IView; Context: Pointer): Boolean; cdecl;

	IEOFileThread = interface(IInterface)
		['{43ACB7AD-DD13-4D76-9A72-0E8F24E5E4A6}']
		function GetActiveView: IView;
		procedure SetActiveView(const Value: IView);
		property ActiveView: IView read GetActiveView write SetActiveView;
		function SelectFolder(const Title: WideString; out Path: TPath): Boolean;
		function Enumerate(Proc: CEachView; Context: Pointer): IView;
	end;

	IEOFileGlobal = interface(IInterface)
		['{2087F0CF-CFE7-4BF2-AD24-1219F0F58176}']
		function GetExePath: WideString;
		function GetInitialFilter: WideString;
		function GetEditor: WideString;
		procedure StartViewThread(const AFirstPath: TPath);
		procedure StartCommandLineThread(const Title, Dir, Initial: WideString; CaretIndex, SelectedLength: Integer);
		function ExecuteCommandLine(const Cmd, Dir: WideString; Wait: Boolean = False): Boolean;
		property ExePath: WideString read GetExePath;
		property InitialFilter: WideString read GetInitialFilter;
		property Editor: WideString read GetEditor;
	end;

	CEachExtension = function(const Extension: IExtension; Context: Pointer): Boolean; cdecl;

	IExtensionManager = interface(IInterface)
		['{226F3891-B8EB-42BE-850A-6B53D79A9BD2}']
		procedure RegisterExtension(const AExtension: IExtension);
		function Enumerate(Proc: CEachExtension; Context: Pointer): IExtension;
	end;

	{for extensions}

	IExtension = interface(IInterface)
		['{06A37D72-465C-4FCE-8767-9A720B986A0B}']
		procedure Initialize;
		procedure Finalize;
	end;

	IExtensionInfo = interface(IInterface)
		['{25D60ADA-4ADF-4326-A8D1-E6B02A9A164F}']
		function GetName: WideString;
		procedure Dialog;
		property Name: WideString read GetName;
	end;

	CEachArchiveKind = function(const Ext: WideString; Context: Pointer): Boolean; cdecl;

	IArchiver = interface(IInterface)
		['{0C56D618-56CB-4254-9282-8DD6005CB81F}']
		function Enumerate(Proc: CEachArchiveKind; Context: Pointer): WideString;
		function Open(const APath: TPath): IFolder;
		procedure Pack(const Archive, SourceDir: WideString; const Sources: array of WideString);
		procedure Unpack(const Archive, DestDir: WideString);
	end;

	IThreadHandler = interface(IInterface)
		['{FCD38D0C-5D8B-423B-AE71-44F61B755201}']
		procedure ThreadBegun(const AThread: IInterface);
		procedure ThreadEnded(const AThread: IInterface);
	end;

	IViewHandler = interface(IInterface)
		['{9ADAD06D-C149-43D8-B182-B5A0AC1A473C}']
		procedure OnOpen(const View: IView);
		procedure OnClose(const View: IView);
	end;

	IPathHandler = interface(IInterface)
		['{50CBFB83-D50A-4C20-A58E-0BEB5901D441}']
		function TryOpen(const APath: TPath): IFolder;
	end;

const
	CoreModule = 'eofile.exe';

function GetApplication: IInterface;
external CoreModule name 'EOFile@GetApplication';

function GetCurrentThread: IInterface;
external CoreModule name 'EOFile@GetCurrentThread';

function GetInterfaceVersion: Cardinal;
external CoreModule name 'EOFile@GetInterfaceVersion';

implementation

end.

