unit UActionAttribute;

{ EOFile by YT }

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
	Windows, Messages, SysUtils, Types,
	SysUtilsW, SysHacks, WinTips,
	LConsts, LObjects, LWindows, LActions, LStdCtls, LAddCtls, LCmmCtls, LCmmDlgs,
	EOFileIntf, UFolders, UView;

type
	TAttributeAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

implementation

type
	TAttr = (attrReadOnly, attrHidden, attrSystem, attrArchive);
	TAttrSwitch = (asPass, asOn, asOff);
	TAttrSwitches = array[TAttr] of TAttrSwitch;

	TTimeStamp = (tsCreation, tsWrite, tsAccess);

	TAttributeDialog = class(T0DialogWindow)
	private
		FSwitches: array[TAttr] of TLightSwitch;
		FTimeStamps: array[TTimeStamp] of record
			Text: TStaticText;
			DateTime: TDateTimePicker;
		end;
		FOk: TButton;
		FCancel: TCloseButton;
		FPass: Boolean;
		function GetSwitches(Attr: Cardinal): TAttrSwitch;
		procedure SetSwitches(Attr: Cardinal; const Value: TAttrSwitch);
		function GetDateTimePickers(Index: TTimeStamp): TDateTimePicker;
	protected
		procedure Initialize; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
	public
		constructor Create(AParent: T0TopWindow; APass: Boolean);
		property Switches[Attr: Cardinal]: TAttrSwitch read GetSwitches write SetSwitches;
		property DateTimePickers[Index: TTimeStamp]: TDateTimePicker read GetDateTimePickers;
	end;

const
	BooleanToSwitch: array[Boolean] of TAttrSwitch = (asOff, asOn);

function fa2Attr(Value: Cardinal): TAttr;
begin
	case Value of
	faReadOnly: Result := attrReadOnly;
	faHidden:   Result := attrHidden;
	faSysFile:  Result := attrSystem;
	faArchive:  Result := attrArchive;
	else
		Assert(False);
		Result := attrReadOnly;
	end;
end;

constructor TAttributeDialog.Create(AParent: T0TopWindow; APass: Boolean);
begin
	inherited Create(AParent, '');
	FPass := APass;
end;

procedure TAttributeDialog.Initialize;
const
	S: array[TAttr] of WideString = ('Read-Only', 'Hidden', 'System', 'Archive');
	TSText: array[TTimeStamp] of WideString = ('Creation', 'Write', 'Access');
var
	FS, SS, C, DS: TSize;
	I: TAttr;
	Item: T0ControlItem;
	T: TTimeStamp;
	PickerHeight: Integer;
	PickerStyle: TDateTimePickerStyles;
begin
	inherited;

	FS := GetFontSize(DefaultGUIFont.Handle, StandardFontWidth);
	PickerHeight := OneLineDefaultSize.cy;

	SS.cx := FS.cx * 30 + 12;
	SS.cy := FS.cy + 8;

	C.cx := SS.cx + 8;
	C.cy := SS.cy * 4 + 36 + PickerHeight * 3;

	DS.cx := 140;
	DS.cy := PickerHeight;

	ClientSize := C;

	for I := Low(I) to High(I) do
	begin
		FSwitches[I] := TLightSwitch.
						Create(Self, S[I], Point(4, Ord(I) * SS.cy), SS);
		FSwitches[I].Font := DefaultGUIFont;
		FSwitches[I].TextWidth := FS.CX * 15;
		FSwitches[I].ItemWidth := FS.CX *  5;

		Item := FSwitches[I].Add;
		Item.Text := 'ON';
		Item.Data := Pointer(asOn);
		FSwitches[I].Selected := Item;

		if FPass then
		begin
			Item := FSwitches[I].Add;
			Item.Text := '-';
			Item.Data := Pointer(asPass);
			FSwitches[I].Selected := Item;
		end;

		Item := FSwitches[I].Add;
		Item.Text := 'OFF';
		Item.Data := Pointer(asOff); 
	end;

	PickerStyle := [];
	if FPass then PickerStyle := [dtsCheckBoxs];

	for T := Low(T) to High(T) do
	begin
		FTimeStamps[T].Text := TStaticText.Create(Self, TSText[T],
			Point(8, Ord(T) * PickerHeight + SS.cy * 4 + 4), DefaultSize);
		FTimeStamps[T].DateTime := TDateTimePicker.Create(Self,
			Point(56, Ord(T) * PickerHeight + SS.cy * 4 + 4), DS, PickerStyle);
		FTimeStamps[T].DateTime.Format := 'yyyy-MM-dd HH:mm:ss';
		if FPass then
      FTimeStamps[T].DateTime.Checked := False;
	end;

	ActiveControl := FSwitches[Low(TAttr)];

	FOK := TButton.Create(Self, SOK, Point(C.CX - 160, C.CY - 28), DefaultSize);
	FOK.IsDefault := True;

	FCancel := TCloseButton.Create(Self, SCancel,
					Point(C.CX - 82, C.CY - 28), DefaultSize);
end;

function TAttributeDialog.GetSwitches(Attr: Cardinal): TAttrSwitch;
var
	R: TLightSwitch;
begin
	R := FSwitches[fa2Attr(Attr)];
	Assert(R.Selected <> nil);
	Result := TAttrSwitch(R.Selected.Data);
end;

procedure TAttributeDialog.
				SetSwitches(Attr: Cardinal; const Value: TAttrSwitch);
var
	R: TLightSwitch;
begin
	R := FSwitches[fa2Attr(Attr)];
	case Value of
	asOn:  R.Selected := R.At[0];
	asOff: R.Selected := R.At[R.Count - 1];
	else   R.Selected := R.At[1];
	end;
end;

procedure TAttributeDialog.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	if (Sender = FOK) and (Info.Code = NM_COMMAND + BN_CLICKED) then
	begin
		DialogResult := ID_OK;
		Close
	end;
	inherited;
end;

{ TAttributeAction }

procedure Switch(var A: Cardinal; Attr: Cardinal; S: TAttrSwitch);
begin
	case S of
	asOn:   A := A or Attr;
	asOff:  A := A and not Attr;
	end;
end;

constructor TAttributeAction.Create;
begin
	inherited Create(False);
	SetName('Attribute');
	SetText('...');
	SetHint('It@C̑ύX܂');
end;

procedure TAttributeAction.DoExecute;
var
	Dialog: TAttributeDialog;
	Multi: Boolean;
	Attr, NewAttr: DWORD;
	View: TFilerView;
	I: Integer;
	Path, FullPathName: WideString;
	CTime, WTime, ATime, LocalTime: FILETIME;
	SystemTime: TSystemTime;
	CurrentFile: PFile;
	H: THandle;
begin
	View := ActiveView;
	Multi := ActiveView.SelectCount > 0;

	Dialog := TAttributeDialog.Create(CurrentThread.LeadWindow, Multi);
	Dialog.Owner := Garbage;

	if not Multi then
	begin
		CurrentFile := View.FocusedFile;

		Attr := CurrentFile^.Attributes;
		Dialog.Switches[faReadOnly] := BooleanToSwitch[Attr and faReadOnly <> 0];
		Dialog.Switches[faHidden]   := BooleanToSwitch[Attr and faHidden <> 0];
		Dialog.Switches[faSysFile]  := BooleanToSwitch[Attr and faSysFile <> 0];
		Dialog.Switches[faArchive]  := BooleanToSwitch[Attr and faArchive <> 0];

		FileTimeToLocalFileTime(CurrentFile^.CreationTime, LocalTime);
		FileTimeToSystemTime(LocalTime, SystemTime);
		Dialog.DateTimePickers[tsCreation].Value := SystemTime;

		FileTimeToLocalFileTime(CurrentFile^.LastWriteTime, LocalTime);
		FileTimeToSystemTime(LocalTime, SystemTime);
		Dialog.DateTimePickers[tsWrite].Value := SystemTime;

		FileTimeToLocalFileTime(CurrentFile^.LastAccessTime, LocalTime);
		FileTimeToSystemTime(LocalTime, SystemTime);
		Dialog.DateTimePickers[tsAccess].Value := SystemTime;
	end;

	if Dialog.Execute then
	begin
		Path := IncludeTrailingPathDelimiter(View.Model.Path.Directory);

		for I := View.ItemCount - 1 downto 0 do
			if View.Selected[I] or (not Multi and (I = View.CaretIndex)) then
			begin
				CurrentFile := View.Items[I];

				with CurrentFile^ do SetString(FullPathName, Name, NameLength);
				FullPathName := Path + FullPathName;

				Attr := GetFileAttributesW(Pointer(FullPathName));
				if Attr = High(Attr) then RaiseLastOSError;
				NewAttr := Attr;
				Switch(NewAttr, faReadOnly, Dialog.Switches[faReadOnly]);
				Switch(NewAttr, faHidden, Dialog.Switches[faHidden]);
				Switch(NewAttr, faSysFile, Dialog.Switches[faSysFile]);
				Switch(NewAttr, faArchive, Dialog.Switches[faArchive]);

				CTime := CurrentFile^.CreationTime;
				if Dialog.DateTimePickers[tsCreation].Checked then
				begin
					SystemTimeToFileTime(Dialog.DateTimePickers[tsCreation].Value, LocalTime);
					LocalFileTimeToFileTime(LocalTime, CTime);
				end;

				WTime := CurrentFile^.LastWriteTime;
				if Dialog.DateTimePickers[tsWrite].Checked then
				begin
					SystemTimeToFileTime(Dialog.DateTimePickers[tsWrite].Value, LocalTime);
					LocalFileTimeToFileTime(LocalTime, WTime);
				end;

				ATime := CurrentFile^.LastAccessTime;
				if Dialog.DateTimePickers[tsAccess].Checked then
				begin
					SystemTimeToFileTime(Dialog.DateTimePickers[tsAccess].Value, LocalTime);
					LocalFileTimeToFileTime(LocalTime, ATime);
				end;

				if (Int64(CTime) <> Int64(CurrentFile^.CreationTime)) or
					(Int64(WTime) <> Int64(CurrentFile^.LastWriteTime)) or
					(Int64(ATime) <> Int64(CurrentFile^.LastAccessTime)) then
				begin
					if (Attr and faReadOnly <> 0) and
						not SetFileAttributesW(Pointer(FullPathName), NewAttr and not faReadOnly) then
					begin
						TMessageBox.Execute(SysErrorMessage(GetLastError), Text, MB_ICONERROR);
						Break;
					end;

					H := CreateFileW(Pointer(FullPathName), GENERIC_READ or GENERIC_WRITE,
						FILE_SHARE_READ or FILE_SHARE_WRITE,
						nil, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0);
					if H = INVALID_HANDLE_VALUE then
					begin
						TMessageBox.Execute(SysErrorMessage(GetLastError), Text, MB_ICONERROR);
						Break;
					end;
					SetFileTime(H, @CTime, @ATime, @WTime);
					CloseHandle(H);
				end;

				if not SetFileAttributesW(Pointer(FullPathName), NewAttr) then
				begin
					TMessageBox.Execute(SysErrorMessage(GetLastError), Text, MB_ICONERROR);
					Break;
				end;
			end;
	end;

	//View.Model.Refresh; m̂ŕsv
end;

procedure TAttributeAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	SetEnabled(View.SelectedIsRealFiles);
end;

function TAttributeDialog.GetDateTimePickers(Index: TTimeStamp): TDateTimePicker;
begin
	Result := FTimeStamps[Index].DateTime
end;

end.
