unit UActionCommandLine;

{ EOFile by YT }

interface

uses
	Windows, Messages, Types, SysUtils,
	SysUtilsW, AlgUtils, WinTips,
	LWindows, LActions, LStdCtls, LCmmDlgs, LAddDlgs,
	EOFileIntf, UView, UFolders;

type
	TCommandLineAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

	TCommandLineThread = class(T0GUIThread)
	private
		FTitle: WideString;
		FDir: WideString;
		FInitial: WideString;
		FCaretIndex: Integer;
		FSelectedLength: Integer;
	protected
		procedure Execute; override;
	public
		constructor Create(const Title, Dir, Initial: WideString; CaretIndex, SelectedLength: Integer);
	end;

implementation

type
	TOuterThread = class(T0Thread)
	private
		FTempFileName: WideString;
	protected
		procedure Execute; override;
	public
		constructor Create(const ATempFileName: WideString);
	end;

constructor TOuterThread.Create(const ATempFileName: WideString);
begin
	FTempFileName := ATempFileName;
	inherited Create;
end;

procedure TOuterThread.Execute;
var
	Editor: WideString;
begin
	Editor := (Application as IEOFileGlobal).Editor;
	if Editor = '' then Editor := 'notepad.exe';

	ExecuteCommandLine('"' + Editor + '" "' + FTempFileName + '"', TemporaryDirectory, SW_NORMAL, True);
	DeleteFileW(PWideChar(FTempFileName));
end;

type
	TCommandLineWindow = class(T0DialogWindow)
	private
		FCommandLine: TLineField;
		FPipe: TCheckBox;
		FOuter: TCheckBox;
		FResult: TTextField;
		FRun: TButton;
		FClose: TCloseButton;
		FDirectory: WideString;
	protected
		procedure Initialize; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		procedure Setup; override;
		property CommandLine: TLineField read FCommandLine;
		property Directory: WideString read FDirectory write FDirectory;
		procedure ExecuteWin;
		procedure ExecutePipe;
	public
		constructor Create(AParent: T0TopWindow; const AText: WideString);
		procedure Realign; override;
		procedure Update; override;
	end;

constructor TCommandLineWindow.Create(AParent: T0TopWindow; const AText: WideString);
var
	CS: PCreateStructW;
begin
	inherited;
	CS := CreateStruct;
	CS^.style := CS^.style or WS_THICKFRAME;
end;

procedure TCommandLineWindow.ExecutePipe;
const
	CmdError = ''' ́AR}h܂͊OR}hA'#13#10 +
		'\ȃvO܂̓ob` t@CƂĔFĂ܂B'#13#10;
var
	Cmd, TempFileName: WideString;
	StdOut, StdErr: string;
	S: TSize;
	ErrorLevel: DWORD;
	H: THandle;
begin
	Cmd := CommandLine.Text;
	try
		try
			SetCurrentDirectoryW(PWideChar(Directory));
			if not ExecuteCommandLine('cmd /c ' + Cmd, Directory, '', StdOut, @StdErr, @ErrorLevel) or
				((ErrorLevel = 1) and CompareMem(PChar(StdErr) + Length(StdErr) - Length(CmdError), PChar(CmdError), Length(CmdError)))
			then
				raise Exception.Create('st@C܂ł');

			StdOut := Directory + '> ' + Cmd + #13#10 + StdOut;
			if StdErr <> '' then
				StdOut := StdOut + #13#10'<<<<<<<< ERROR OUTPUT >>>>>>>>'#13#10 + StdErr;

			if FOuter.Checked then
			begin
				TempFileName := WinTips.TemporaryFileName;
				H := FileCreate(TempFileName);
				FileWrite(H, Pointer(StdOut)^, Length(StdOut));
				FileClose(H);
				TOuterThread.Create(TempFileName);
        Close
			end
			else
			begin
				FResult.Text := StdOut;
				S := BoundsSize;
				if S.CY < 100 then
					Inc(S.CY, 250 + 4);
				BoundsSize := S;
			end
		finally
			SetCurrentDirNull
		end;
	except
	on E: Exception do
		TMessageBox.Execute(E.Message, Cmd, MB_ICONERROR);
	end;
end;

procedure TCommandLineWindow.ExecuteWin;
var
	Cmd: WideString;
begin
	Cmd := CommandLine.Text;
	try
		try
			SetCurrentDirectoryW(PWideChar(Directory));
			if not ExecuteCommandLine(Cmd, Directory) then
				raise Exception.Create('st@C܂ł');
			Close;
		finally
			SetCurrentDirNull
		end;
	except
	on E: Exception do
		TMessageBox.Execute(E.Message, Cmd, MB_ICONERROR);
	end;
end;

procedure TCommandLineWindow.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	if (Sender = FRun) and (Info.Code = NM_COMMAND + BN_CLICKED) then
	begin
		if FPipe.Checked then
			ExecutePipe
		else
			ExecuteWin
	end
	else
		inherited;
end;

procedure TCommandLineWindow.Initialize;
const
	LS: TSize = (cx: 400; cy: UseDefault);
var
	P: TPoint;
begin
	FCommandLine := TLineField.Create(Self, '', Point(2, 4), LS);
	P := FCommandLine.BoundsRect.BottomRight;
	Inc(P.Y, 4);
	FPipe := TCheckBox.Create(Self, 'o͂D(&S)', Point(8, P.Y), DefaultSize);
	FOuter := TCheckBox.Create(Self, 'OGfB^(&O)', Point(128, P.Y), DefaultSize);
	Dec(P.X, 75 * 2 + 2 + 4);
	FRun := TButton.Create(Self, 'Run', P, DefaultSize);
	FRun.IsDefault := True;
	Inc(P.X, 75 + 4);
	FClose := TCloseButton.Create(Self, 'Close', P, DefaultSize);
	FResult := TTextField.Create(Self, '', ZeroPoint, ZeroSize);
	FResult.ReadOnly := True;
	FResult.Font := FixedFont;
	inherited;
end;

procedure TCommandLineWindow.Realign;
var
	R, LR: TRect;
	P: TPoint;
begin
	if ActiveControl <> nil then
	begin
		R := ClientRect;
		R.Top := FClose.BoundsRect.Bottom + 4;
		FResult.BoundsRect := R;

		LR := FCommandLine.BoundsRect;
		LR.Right := R.Right - 2;
		FCommandLine.BoundsRect := LR;

		P := FClose.Position;
		P.X := LR.Right - 79;
		FClose.Position := P;

		P.X := LR.Right - 79 * 2;
		FRun.Position := P;
	end
end;

procedure TCommandLineWindow.Setup;
var
	S: TSize;
begin
	TPoint(S) := FClose.BoundsRect.BottomRight;
	Inc(S.cx, 4);
	Inc(S.cy, 2);
	ClientSize := S;
	ActiveControl := FCommandLine;
	inherited;
end;

procedure TCommandLineWindow.Update;
begin
	FOuter.Enabled := FPipe.Checked;
	inherited;
end;

{ TCommandLineThread }

constructor TCommandLineThread.Create(const Title, Dir, Initial: WideString; CaretIndex, SelectedLength: Integer);
begin
	FTitle := Title;
	FDir := Dir;
	FInitial := Initial;
	FCaretIndex := CaretIndex;
	FSelectedLength := SelectedLength;
	inherited Create;
end;

procedure TCommandLineThread.Execute;
var
	Hidden: HWND;
	Dialog: TCommandLineWindow;
begin
	Hidden := CreateWindowExW(0, 'STATIC', '', 0, 0, 0, 0, 0, 0, 0, MainInstance, nil);
	Application.BeginRead;
	try
		Dialog := TCommandLineWindow.Create(nil, FTitle);
		Dialog.Owner := Garbage;
		SetWindowLong(Dialog.Handle, GWL_HWNDPARENT, Hidden);
		Dialog.Directory := FDir;
		Dialog.CommandLine.Text := FInitial;
		Dialog.CommandLine.CaretIndex := FCaretIndex;
		Dialog.CommandLine.SelectedLength := FSelectedLength;
		Dialog.Execute;
	finally
		Application.EndRead;
		DestroyWindow(Hidden);
	end;
end;

{ TCommandLineAction }

constructor TCommandLineAction.Create;
begin
	inherited Create(False);

	SetName('CommandLine');
	SetText('R}hC...');
	SetHint('R}hCss܂');
end;

procedure TCommandLineAction.DoExecute;
const
	ExeExts: array[0..3] of String = ('.exe', '.com', '.bat', '.cmd');
var
	View: TFilerView;
	I, CaretIndex, SelLength: Integer;
	Cmd, W, E: WideString;
	Exe: Boolean;
	Focused: PFile;
begin
	View := ActiveView; {eXbh}
	if View.SelectCount > 0 then
	begin
		Cmd := '';
		for I := View.ItemCount - 1 downto 0 do
		begin
			if View.Selected[I] then
			begin
				Focused := View.Items[I];
				SetString(W, Focused.Name, Focused.NameLength);
				if Pos(#32, W) > 0 then W := '"' + W + '"';
				Cmd := #32 + W + Cmd;
			end;
		end;
		CaretIndex := 0;
		SelLength := 0;
	end
	else if View.ItemCount > 0 then
	begin
		Focused := View.FocusedFile;
		SetString(W, Focused.Name, Focused.NameLength);
		E := ExtractFileExt(W);
		Exe := False;
		for I := Low(ExeExts) to High(ExeExts) do
			Exe := Exe or (CompareFileName(E, ExeExts[I]) = 0);

		if Pos(#32, W) > 0 then W := '"' + W + '"';

		if Exe then
		begin
			Cmd := W + ' ';
			CaretIndex := Length(Cmd);
			SelLength := 0;
		end
		else
		begin
			Cmd := ' ' + W;
			CaretIndex := 0;
			SelLength := 0;
		end;
	end
	else
	begin
		Cmd := '';
		CaretIndex := 0;
		SelLength := 0;
	end;

	TCommandLineThread.Create('R}hCs', View.Model.Path.Directory, Cmd, CaretIndex, SelLength);
end;

procedure TCommandLineAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	SetEnabled((View <> nil) and Supports(View.Model, IDirectory));
end;

end.
