unit UActionDelete;

{ EOFile by YT }

interface

uses
	Windows, SysUtils, ShellApi,
	SysUtilsW, AlgUtils, ShlTips,
	LConsts, LActions, LWindows, LAddDlgs,
	EOFileIntf, UPath, UFolders, UView;

type
	TDeleteAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

implementation

type
	TDeleteThread = class(T0Thread)
	private
		FBasePath: WideString;
		FFiles: TWideFileNameArray;
		FFlag: FILEOP_FLAGS;
	protected
		procedure Execute; override;
	public
		constructor Create(const BasePath: WideString; const Files: TWideFileNameArray; Flag: FILEOP_FLAGS);
	end;

{ TDeleteThread }

constructor TDeleteThread.Create(const BasePath: WideString; const Files: TWideFileNameArray; Flag: FILEOP_FLAGS);
begin
  FBasePath := BasePath;
	FFiles := Files;
	FFlag := Flag;
	inherited Create;
end;

procedure TDeleteThread.Execute;
var
	I: Integer;
	DeletedPath: TPath;
	FolderManager: IFolderManager;
begin
	FolderManager := Application as IFolderManager;

	try
		ShellDeleteFiles(0, FBasePath, FFiles, FFlag);
	except
	on E: Exception do
		ShowException(ExceptObject, ExceptAddr);
	end;

	Application.BeginWrite;
	try
		InitializePath(DeletedPath);
		for I := Low(FFiles) to High(FFiles) do
		begin
			DeletedPath.Directory := FBasePath + FFiles[I];
			FolderManager.Refresh(DeletedPath, True);
		end;
	finally
		Application.EndWrite;
	end
end;

{ TDeleteAction }

constructor TDeleteAction.Create;
begin
	inherited Create(False);
	SetName('Delete');
	SetText('폜...');
	SetHint('t@C폜܂');
end;

procedure TDeleteAction.DoExecute;
var
	View: TFilerView;
	Dialog: TSelectDialog;
	FileNames: TFilePtrArray;
	BasePath: WideString;
	Flag: FILEOP_FLAGS;
begin
	View := ActiveView;

	Dialog := TSelectDialog.Create(CurrentThread.LeadWindow, '폜');
	Dialog.Owner := Garbage;
	with Dialog.Add do begin Data := Pointer(FOF_ALLOWUNDO); Text := 'ݔ'; end;
	with Dialog.Add do begin Data := Pointer(0); Text := ''; end;

	if Dialog.Execute then
	begin
		BasePath := IncludeTrailingPathDelimiter(View.Model.Path.Directory);
		FileNames := View.SelectedFiles;
		Flag := FILEOP_FLAGS(Dialog.Selected.Data);

		TDeleteThread.Create(BasePath, FilePtrArray2WideFileNameArray(FileNames), Flag);
	end;
	FileNames := nil;
end;

procedure TDeleteAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	SetEnabled(View.SelectedIsRealFiles);
end;

end.

