unit UActionFindFile;

{ EOFile by YT }

interface

uses
	Windows, Messages, SysUtils, Types,
	SysUtilsW, WinTips, DYARegExp,
	LConsts, LObjects, LGraphics, LActions, LWindows,
	LStdCtls, LAddCtls, LCmmDlgs,
	EOFileIntf, UView, UFolders, UPath;

type
	TFindFileAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

	TFounds = class(T0Folder, IFolder, IFounds)
	private
		FCount: Integer;
		FPreFiles: TFileArray;
		FPreUpdated: DWORD;
		FBreakEvent: THandle;
		FWildCard: WideString;
		{IFolder}
		function GetDisplayText: WideString;
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		procedure Move(const NewPath: WideString);
		function GetInfoText(AFile: PFile): WideString;
		function GetPossibilities: TFolderPossibilities;
		function ConcatenatePath(AFile: PFile): TPath;
		{IFounds}
		procedure Add(const AFile: TFile);
		procedure Fin;
		function GetWildCard: WideString;
	protected
		procedure DoRefresh; override;
		procedure ViewDecreasedAll; override;
	public
		constructor Create(const APath: TPath; const AWild: TWildCardExpression; const ARegExp: IRegExp);
		procedure BeforeDestruction; override;
	end;

implementation

type
	TGrepThread = class(T0Thread)
	private
		FFounds: TFounds;
		FRegExp: IRegExp;
		FWild: TWildCardExpression;
		FBreak: THandle;
		FPath: WideString;
	protected
		procedure Execute; override;
		procedure Finalize; override;
	public
		constructor Create(const AFounds: TFounds; const APath: WideString;
						AWild: TWildCardExpression; const ARegExp: IRegExp; ABreak: THandle);
	end;

constructor TGrepThread.Create(const AFounds: TFounds;
	const APath: WideString; AWild: TWildCardExpression; const ARegExp: IRegExp; ABreak: THandle);
begin
	FFounds := AFounds;
	FFounds._AddRef;

	FWild := AWild;
	FBreak := DuplicateHandle(ABreak);
	FRegExp := ARegExp;
	FPath := APath;
	inherited Create;
end;

procedure TGrepThread.Execute;

	procedure Search(const Path: WideString);
	label
		Dir;
	var
		FindData: WIN32_FIND_DATAW;
		H, F: THandle;
		Mask, FullPathName: TWideFileName;
		Data: AnsiString; {binary}
		P: TFile;
	begin
		Mask := Path + '*';

		H := FindFirstFileW(PWideChar(Mask), FindData);
		if H <> INVALID_HANDLE_VALUE then
			try
				repeat
					if WaitForSingleObject(FBreak, 0) = WAIT_OBJECT_0 then Abort;

					if (FindData.cFileName[0] <> '.') or
									(FindData.cFileName[1] <> '.') and (FindData.cFileName[1] <> #0) then
					begin
						if FWild.Evaluate(FindData.cFileName, FindData.dwFileAttributes) then
						begin
							FullPathName := Path + FindData.cFileName;

							if FRegExp <> nil then
							begin
								if FindData.dwFileAttributes and faDirectory <> 0 then goto Dir;
								F := SysUtilsW.FileOpen(FullPathName, fmOpenRead or fmShareDenyWrite);
								if F = INVALID_HANDLE_VALUE then goto Dir;
								SetLength(Data, GetFileSize(F));
								FileRead(F, Pointer(Data)^, Length(Data));
								FileClose(F);
								if not MatchExists(FRegExp, Data) then goto Dir;
							end;

							FindDataToFile(P, FindData, Application as IFolderManager);
							if P.ShortName[0] = #0 then
							begin
								lstrcpyW(P.ShortName, P.Name);
								P.ShortNameLength := P.NameLength;
							end;
							lstrcpyW(P.Name, PWideChar(FullPathName));
							P.NameLength := Length(FullPathName);
							FFounds.Add(P);
						end;

						Dir: if FindData.dwFileAttributes and faDirectory <> 0 then
						begin
							Search(Path + FindData.cFileName + '\');
						end;
					end;
				until not FindNextFileW(H, FindData);
			finally
				Windows.FindClose(H);
			end;
	end;

begin
	Search(IncludeTrailingPathDelimiter(FPath));
	FFounds.Fin;
end;

procedure TGrepThread.Finalize;
begin
	FWild.Finalize;
	FFounds._Release;
	CloseHandle(FBreak);
	inherited;
end;

type
	TGrepWindow = class(T0DialogWindow)
	private
		FWildCard, FRegExp: TLineField;
		FStart: TButton;
		FCancel: TCloseButton;
		FWildCardObj: TWildCardExpression;
		FRegExpObj: IRegExp;
	protected
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		procedure Setup; override;
		procedure PaintWindow(DC: HDC); override;
	end;

procedure TGrepWindow.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
var
	Error: TWildCardErrorKind;
begin
	if (Sender = FStart) and (Info.Code = NM_COMMAND + BN_CLICKED) then
	begin
		try
			FWildCardObj.Initialize(FWildCard.Text, Error);
			if Error <> weNone then
			begin
				TMessageBox.Execute('ChJ[hԈႢ', '', MB_ICONERROR);
				FWildCard.SetFocus;
			end
			else
			begin
				FRegExpObj := nil;
				if FRegExp.Text <> '' then FRegExpObj := RegExp(FRegExp.Text, reeLocale, [reoCaseInsensitive]);

				DialogResult := ID_OK;
				Close
			end
		except
		on EConvertError do
			begin
				TMessageBox.Execute('K\ԈႢ', '', MB_ICONERROR);
				FRegExp.SetFocus
			end;
		end;
	end;
	inherited;
end;

procedure TGrepWindow.PaintWindow(DC: HDC);
begin
	SelectObject(DC, DefaultGUIFont.Handle);
	SetBkMode(DC, TRANSPARENT);
	TextOut(DC, 4,  8, 't@C');
	TextOut(DC, 4, 32, 'g(K\)');
end;

procedure TGrepWindow.Setup;
var
	R: TRect;
	Height, Width, Base: Integer;
begin
	ClientSize := Size(320, 28 + 24 + 24 + 4);
	FWildCard := TLineField.Create(Self, '"*.*"', Point(100, 4), DefaultSize);
	FRegExp := TLineField.Create(Self, '', Point(100, 28), DefaultSize);
	ActiveControl := FWildCard;
	FWildCard.CaretIndex := 1;
	FWildCard.SelectedLength := 3;

	FStart := TButton.Create(Self, 'Start', ZeroPoint, DefaultSize);
	FStart.IsDefault := True;
	FCancel := TCloseButton.Create(Self, SCancel, ZeroPoint, DefaultSize);

	R := ClientRect;
	Height := FWildCard.BoundsSize.CY;
	Width := R.Right - 4 - FWildCard.BoundsRect.Left;
	FWildCard.BoundsSize := Size(Width, Height);
	FRegExp.BoundsSize := Size(Width, Height);

	Base := FRegExp.BoundsRect.Bottom + 5;

	FStart.Position := Point(R.Right - 150 - 12, Base);
	FCancel.Position := Point(R.Right - 75 - 8, Base);

	inherited Setup;
end;

{ TFindFileAction }

constructor TFindFileAction.Create;
begin
	inherited Create(False);
	SetName('FindFile');
	SetText('t@C...');
	SetHint('t@C܂');
end;

procedure TFindFileAction.DoExecute;
var
	Dialog: TGrepWindow;
	Founds: IFolder;
	View: TFilerView;
begin
	Dialog := TGrepWindow.Create(T0GUIThread(CurrentThread).LeadWindow, 't@C');
	Dialog.Owner := Garbage;
	if Dialog.Execute then
	begin
		View := ActiveView;
		Founds := TFounds.Create(View.Model.Path, Dialog.FWildCardObj, Dialog.FRegExpObj);
		View.Model := Founds;
	end;
end;

procedure TFindFileAction.Update;
begin
	SetEnabled(Supports(ActiveView.Model, IDirectory) or Supports(ActiveView.Model, IFounds));
end;

{ TFounds }

procedure TFounds.Add(const AFile: TFile);
var
	N: DWORD;
begin
	Application.BeginWrite;

	if FCount = Length(FFiles) then
		SetLength(FFiles, FCount + 1024);

	FFiles[FCount] := AFile;
	Inc(FCount);

	N := GetTickCount;
	if N > FPreUpdated + 1000 then
	begin
		FPreUpdated := N;
		SetLength(FFiles, FCount);
		Views.Notify;
		FPreFiles := FFiles;
	end;

	Application.EndWrite;
end;

procedure TFounds.BeforeDestruction;
begin
	CloseHandle(FBreakEvent);
	inherited;
end;

function TFounds.ConcatenatePath(AFile: PFile): TPath;
begin
	InitializePath(Result);
	Result.Directory  := ConcatPath(Path.Directory, AFile.Name);
end;

constructor TFounds.Create(const APath: TPath; const AWild: TWildCardExpression; const ARegExp: IRegExp);
begin
	inherited Create(APath);
	FValid := True;
	FBreakEvent := CreateEventW(nil, True, False, nil);
	FPreUpdated := GetTickCount;
	FWildCard := AWild.Source;
	TGrepThread.Create(Self, Path.Directory, AWild, ARegExp, FBreakEvent);
end;

procedure TFounds.DoRefresh;
begin
end;

procedure TFounds.Fin;
begin
	Application.BeginWrite;
	SetLength(FFiles, FCount);
	Views.Notify;
	FPreFiles := nil;
	Application.EndWrite
end;

function TFounds.GetDisplayText: WideString;
begin
	Result := Path.Directory
end;

function TFounds.GetInfoText(AFile: PFile): WideString;
begin
	if AFile <> nil then
		SetString(Result, AFile.Name, AFile.NameLength)
	else
		Result := ''
end;

function TFounds.GetParentPath: TPath;
begin
	Result := CopyPath(Path)
end;

function TFounds.GetPossibilities: TFolderPossibilities;
begin
	Result := [fpSort, fpHasParent]
end;

function TFounds.GetWildCard: WideString;
begin
	Result := FWildCard
end;

function TFounds.IsSub(const AParent: TPath): Boolean;
begin
	Result := DirectoryIsSub(Path.Directory, AParent)
end;

procedure TFounds.Move(const NewPath: WideString);
begin
	FValid := False;
	SetEvent(FBreakEvent);
end;

procedure TFounds.ViewDecreasedAll;
begin
	SetEvent(FBreakEvent);
	inherited;
end;

end.
