unit UActionIncrementalSearch;

{ EOFile by YT }

interface

uses
	Windows, Messages, SysUtils, Types, ActiveX, ShellApi, ShlObj,
	SysUtilsW, WinTips,
	LObjects, LWindows, LActions,
	EOFileIntf, UPath, UFolders, UView;

type
	TIncrementalSearchAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

implementation

type
	TIncrementalSearchDialog = class(T0DialogWindow)
	private
		FFinding: WideString;
		FBasePath: TPath;
	protected
		procedure Setup; override;
		function KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean; override;
		function KeyInput(Code: WideChar): Boolean; override;
		procedure PaintWindow(DC: HDC); override;
		procedure Move;
		class procedure CreateClass(var WndClass: TWNDCLASSEXW); override;
		procedure Finalize; override;
	end;

{ TIncrementalSearchDialog }

class procedure TIncrementalSearchDialog.CreateClass(var WndClass: WNDCLASSEXW);
begin
	inherited;
	WndClass.hbrBackground := GetSysColorBrush(COLOR_INFOBK)
end;

procedure TIncrementalSearchDialog.Finalize;
begin
	FinalizePath(FBasePath);
	inherited;
end;

function TIncrementalSearchDialog.KeyDown(ShiftState: TShiftKeys; Code: Byte): Boolean;
begin
	case Code of
	VK_UP:
		if ActiveView.CaretIndex > 0 then
		begin
			ActiveView.CaretIndex := ActiveView.CaretIndex - 1;
			FFinding := ActiveView.FocusedFile.DisplayName;
			Repaint;
			Move
		end;
	VK_DOWN:
		if ActiveView.CaretIndex < ActiveView.ItemCount - 1 then
		begin
			ActiveView.CaretIndex := ActiveView.CaretIndex + 1;
			FFinding := ActiveView.FocusedFile.DisplayName;
			Repaint;
			Move
		end;
	VK_INSERT:
		if ActiveView.ItemCount > 0 then
		begin
			FFinding := ActiveView.FocusedFile.DisplayName;
			Repaint;
		end;
	end;
	Result := True;
end;

function TIncrementalSearchDialog.KeyInput(Code: WideChar): Boolean;
var
	View: TFilerView;
	WorkPath: TPath;
	C, I: Integer;
	Data: HGLOBAL;
	Model: IFolder;
	ActionManager: IActionManager;
	FolderManager: IFolderManager;
begin
	InitializePath(WorkPath);

	ActionManager := CurrentThread as IActionManager;
	FolderManager := Application as IFolderManager;
	View := ActiveView;

	case Code of
	#9, #13:
		Close;
	#8:
	begin
		if FFinding = '' then
		begin
			if ActionManager.Actions['ToParent'].Execute then
				FFinding := View.FocusedFile.DisplayName;
		end
		else if FFinding = '\\' then
			FFinding := ''
		else
			Delete(FFinding, Length(FFinding), 1);
	end;
	#27:
	begin
		if (FFinding = '') and ComparePath(FBasePath, View.Model.Path) then
			Close
		else
		begin
			View.Model := FolderManager.GetFolder(FBasePath);
			FFinding := ''
		end;
	end;
	^V:
	begin
		OpenClipboard(Handle);
		Data := GetClipboardData(CF_UNICODETEXT);
		if Data = 0 then
			Beep
		else
		begin
			WorkPath.Directory := PWideChar(GlobalLock(Data));
			GlobalUnlock(Data);
			CloseClipboard;

			Model := FolderManager.GetFolder(WorkPath);
			if Model.Valid then
			begin
				View.Model := Model;
				FFinding := ''
			end
			else
			begin
				Model := GetParentFolder(Model);
				if Model.Valid then
				begin
					View.Model := Model;
					FFinding := ExtractFileName(WorkPath.Directory);
					View.FocusNearWithDisplayName(FFinding);
				end
				else
					Beep
			end
		end
	end;
	':':
		if Length(FFinding) = 1 then
		begin
			WorkPath.Directory := WideUpCase(FFinding[1]) + ':\';

			View.Model := FolderManager.GetFolder(WorkPath);
			FFinding := ''
		end;
	'\':
	begin
		if FFinding = '' then
		begin
			ActionManager.Actions['ToRoot'].Execute;
			FFinding := '\'
		end
		else if FFinding = '\' then
		begin
			SHGetSpecialFolderLocation(0, CSIDL_NETWORK, WorkPath.ItemIDList);
			try
				View.Model := FolderManager.GetFolder(WorkPath);
			finally
				FinalizePath(WorkPath);
			end;
			FFinding := '\\'
		end
		else if FFinding[1] = '\' then
		begin
			C := 0;
			for I := 1 to Length(FFinding) do
				if FFinding[I] = '\' then
					Inc(C);

			if C = 2 then
			begin
				if Length(FFinding) > 2 then
				begin
					WorkPath := ParsePath(FFinding);
					if WorkPath.ItemIDList = nil then
						Beep
					else
						try
							Model := FolderManager.GetFolder(WorkPath);
							if Model.Valid then
							begin
								View.Model := Model;
								FFinding := '' //FFinding + '\'
							end
							else
								Beep
						finally
							FinalizePath(WorkPath);
						end;
				end
			end
			else if C = 3 then
			begin
				WorkPath.Directory := View.FocusedFile.Name;
				Model := FolderManager.GetFolder(WorkPath);
				if Model.Valid then
				begin
					View.Model := Model;
					FFinding := ''
				end
				else
					Beep;
			end
		end
		else if FFinding = '.' then
		begin
			FFinding := ''
		end
		else if FFinding = '..' then
		begin
			ActionManager.Actions['ToParent'].Execute;
			FFinding := ''
		end
		else
		begin
			ActionManager.Actions['EnterFolder'].Execute;
			FFinding := ''
		end
	end;
	else
		if FFinding = '\' then FFinding := '';
		FFinding := FFinding + Code;
	end;
	Repaint;
	View.FocusNearWithDisplayName(FFinding);
	Move;
	Result := True;
end;

procedure TIncrementalSearchDialog.Move;
var
	R: TRect;
begin
	R := ActiveView.ItemRect(ActiveView.CaretIndex);
	Dec(R.Top, BoundsSize.CY);
	Position := ActiveView.ClientToScreen(R.TopLeft)
end;

procedure TIncrementalSearchDialog.PaintWindow(DC: HDC);
begin
	SetTextColor(DC, GetSysColor(COLOR_INFOTEXT));
	SelectObject(DC, Font.Handle);
	SetBkMode(DC, TRANSPARENT);
	TextOut(DC, 4 - GetSystemMetrics(SM_CXBORDER), 4, FFinding)
end;

procedure TIncrementalSearchDialog.Setup;
var
	CY: Integer;
begin
	inherited;
	Font := ActiveView.Font;
	ClientSize := Size(0, Abs(Font.Height) + 8);
	CY := BoundsSize.cy;
	BoundsSize := Size(ActiveView.ClientSize.CX, CY);
	Position := ActiveView.ClientToScreen(Point(0, -CY));

	FBasePath := CopyPath(ActiveView.Model.Path);
end;

{ TIncrementalSearchAction }

constructor TIncrementalSearchAction.Create;
begin
	inherited Create(False);
	SetName('IncrementalSearch');
	SetText('CN^T[`');
	SetHint('݂̃r[疼OŃt@CT܂');
end;

procedure TIncrementalSearchAction.DoExecute;
var
	Dialog: TIncrementalSearchDialog;
begin
	Dialog := TIncrementalSearchDialog.Create(CurrentThread.LeadWindow, Text);
	Dialog.Owner := Garbage;
	Dialog.Execute
end;

procedure TIncrementalSearchAction.Update;
begin
end;

end.
