unit UActionPack;

{ EOFile by YT }

interface

uses
	Windows, SysUtils,
	SysUtilsW, WinTips, 
	LConsts, LWindows, LActions, LCmmDlgs, LAddDlgs,
	EOFileIntf, UPath, UView, UFolders;

type
	TPackAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

	TUnpackAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

implementation

type
	TPackThread = class(T0Thread)
	private
		FArcFile: WideString;
		FFromPath: WideString;
		FFiles: TWideFileNameArray;
		FArchiver: IArchiver;
	protected
		procedure Execute; override;
	public
		constructor Create(const AArcFile, AFromPath: WideString; const AFiles: TWideFileNameArray; const AArchiver: IArchiver);
	end;

constructor TPackThread.Create(const AArcFile, AFromPath: WideString; const AFiles: TWideFileNameArray; const AArchiver: IArchiver);
begin
	FArcFile := AArcFile;
	FFromPath := AFromPath;
	FFiles := AFiles;
	FArchiver := AArchiver;
	inherited Create;
end;

procedure TPackThread.Execute;
var
	//S: Integer;
	Path: TPath;
	Model: IFolder;
	Manager: IFolderManager;
begin
	Manager := Application as IFolderManager;
	InitializePath(Path);
	Path.Directory := ExtractFileDir(FArcFile);

	//S := Application.BeginUpdate;
	Model := Manager.GetOpenedFolder(Path);
	//Application.EndUpdate(S);
	if Model <> nil then Model.StopSurveillance;
	try
		FArchiver.Pack(FArcFile, FFromPath, FFiles);
	except
	on E: Exception do
		TMessageBox.Execute(E.Message, 'k', MB_ICONERROR)
	end;
	if Model <> nil then Model.ResumeSurveillance(False);
end;

{ TPackAction }

constructor TPackAction.Create;
begin
	inherited Create(False);
	SetName('Pack');
	SetText('k...');
	SetHint('ɂ쐬܂');
end;

procedure TPackAction.DoExecute;
var
	ToPath: TPath;
	View: TFilerView;
	Dialog: TInputDialog;
	Files: TFilePtrArray;
	Manager: IFolderManager;
	Archiver: IArchiver;
	ArcFile: WideString;
	Today: SYSTEMTIME;
	Buffer: array[0..255] of WideChar;
begin
	if SelectFolder('k', ToPath) then
	begin
		Manager := Application as IFolderManager;
		View := ActiveView;
		Files := View.SelectedFiles;

		Dialog := TInputDialog.Create(CurrentThread.LeadWindow, 'k', False);
		Dialog.Owner := Garbage;

		if View.SelectCount = 0 then
			Dialog.Value := Files[0].Name + '.zip'
		else
		begin
			GetLocalTime(Today);
			GetDateFormatW(LOCALE_USER_DEFAULT, DATE_SHORTDATE, @Today, nil, Buffer, 255);
			Dialog.Value := Buffer + '.zip';
		end;

		Dialog.SelectAll;

		if Dialog.Execute then
		begin
			ArcFile := ConcatPath(ToPath.Directory, Dialog.Value);
			Archiver := Manager.FindArchiver(ArcFile);
			if Archiver = nil then
				TMessageBox.Execute('sȊgqł', 'k', MB_ICONWARNING)
			else
			begin
				TPackThread.Create(ArcFile, View.Model.Path.Directory,
					FilePtrArray2WideFileNameArray(Files), Archiver);
			end
		end;
		FinalizePath(ToPath);
	end;
	Files := nil
end;

procedure TPackAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	SetEnabled(View.SelectedIsRealFiles and Supports(View.Model, IDirectory));
end;

type
	TUnpackThread = class(T0Thread)
	private
		FPath: WideString;
		FArcFile: WideString;
	protected
		procedure Execute; override;
	public
		constructor Create(const APath, AArcFile: WideString);
	end;

constructor TUnpackThread.Create(const APath, AArcFile: WideString);
begin
	FPath := APath;
	FArcFile := AArcFile;
	inherited Create;
end;

procedure TUnpackThread.Execute;
var
	Manager: IFolderManager;
	Archiver: IArchiver;
	Model: IFolder;
	Path: TPath;
	//S: Integer;
begin
	Manager := Application as IFolderManager;
	Archiver := Manager.FindArchiver(FArcFile);
	Assert(Archiver <> nil);

	InitializePath(Path);
	Path.Directory := FPath;

	//S := Application.BeginUpdate;
	Model := Manager.GetOpenedFolder(Path);
	//Application.EndUpdate(S);

	if Model <> nil then Model.StopSurveillance;
	try
		Archiver.Unpack(FArcFile, FPath);
	except
	on E: Exception do
		TMessageBox.Execute(E.Message, 'WJ', MB_ICONERROR)
	end;
	if Model <> nil then Model.ResumeSurveillance(False);
end;

{ TUnpackAction }

constructor TUnpackAction.Create;
begin
	inherited Create(False);
	SetName('Unpack');
	SetText('WJ...');
	SetHint('ɂWJ܂');
end;

procedure TUnpackAction.DoExecute;
var
	ToPath: TPath;
	View: TFilerView;
	ArcFile: WideString;
begin
	if SelectFolder('WJ', ToPath) then
	begin
		View := ActiveView;

		with View.FocusedFile^ do SetString(ArcFile, Name, NameLength);
		ArcFile := ConcatPath(View.Model.Path.Directory, ArcFile);

		TUnpackThread.Create(ToPath.Directory, ArcFile);
		FinalizePath(ToPath);
	end;
end;

procedure TUnpackAction.Update;
var
	View: TFilerView;
	TargetFile: PFile;
	E: Boolean;
begin
	View := ActiveView;
	E := View.ItemCount > 0;
	if E then
	begin
		TargetFile := View.FocusedFile;
		E := (TargetFile.Kind = fkFile) and
					(TargetFile^.Attributes and faDirectory = 0) and
					((Application as IFolderManager).FindArchiver(TargetFile^.Name) <> nil);
	end;
	SetEnabled(E);
end;

end.
