unit UActionRename;

{ EOFile by YT }

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
	Windows, Messages, SysUtils, Types, ShellApi,
	SysUtilsW, WinTips, ShlTips,
	LConsts, LObjects, LWindows, LActions, LStdCtls, LAddDlgs, LCmmDlgs,
	EOFileIntf, UPath, UFolders, UView;

type
	TRenameAction = class(T0Action)
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

implementation

type
	TRenameDialog = class(T0DialogWindow)
	private
		FFileName: TLineField;
		FFileExt: TLineField;
		FOk: TButton;
		FCancel: TCloseButton;
	protected
		procedure Initialize; override;
		procedure Setup; override;
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
	public
		property FileName: TLineField read FFileName;
		property FileExt: TLineField read FFileExt;
	end;

procedure TRenameDialog.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	inherited;
	if (Sender = FOK) and (Info.code = BN_CLICKED + NM_COMMAND) then
	begin
		DialogResult := ID_OK;
		Close;
	end;
end;

procedure TRenameDialog.Initialize;
const
	Point_4_4: TPoint = (X: 4; Y: 4);
	Size_200_D: TSize = (CX: 200; CY: UseDefault);
	Point_208_4: TPoint = (X: 208; Y: 4);
	Size_100_D: TSize = (CX: 100; CY: UseDefault);
begin
	inherited;

	FFileName := TLineField.Create(Self, '', Point_4_4, Size_200_D);
	FFileExt := TLineField.Create(Self, '', Point_208_4, Size_100_D);

	FOK := TButton.Create(Self, SOK, ZeroPoint, DefaultSize);
	FOK.IsDefault := True;

	FCancel := TCloseButton.Create(Self, SCancel, ZeroPoint, DefaultSize);
end;

procedure TRenameDialog.Setup;
var
	Y: Integer;
begin
	inherited;

	Y := FFileName.BoundsRect.Bottom + 4;

	FOk.Position := Point(312 - 160, Y);
	FCancel.Position := Point(312 - 82, Y);

	ClientSize := Size(312, Y + 29);

	FFileName.ReadOnly := FFileName.Text = '*';
	FFileExt.ReadOnly := FFileExt.Text = '*';

	FFileName.SelectAll;
	FFileExt.SelectAll;

	if FFileName.ReadOnly then
		ActiveControl := FFileExt
	else
    ActiveControl := FFileName
end;


{ TRenameAction }

constructor TRenameAction.Create;
begin
	inherited Create(False);
	SetName('Rename');
	SetText('O̕ύX...');
	SetHint('t@Ct܂');
end;

procedure TRenameAction.DoExecute;

	procedure DisassembleFileName(const FileName: WideString; var NamePart, ExtPart: WideString);
	begin
		ExtPart := ExtractFileExt(FileName);
		NamePart := Copy(FileName, 1, Length(FileName) - Length(ExtPart));
		Delete(ExtPart, 1, 1);
	end;

	procedure NotifyRename(const OldName, NewName: WideString);
	var
		P: TPath;
	begin
		Application.BeginWrite;
		try
			InitializePath(P);
			P.Directory := OldName;
			(Application as IFolderManager).ChangePath(P, NewName);
		finally
			Application.EndWrite;
		end
	end;

var
	View: TFilerView;
	Dialog: TInputDialog;
	Drive, BasePath, OldName, NewName, NamePart, ExtPart, N2, E2: WideString;
	Target: PFile;
	I: Integer;
	RenameDialog: TRenameDialog;
	Files: TFilePtrArray;
	R: Boolean;
begin
	View := ActiveView;

	Dialog := TInputDialog.Create(CurrentThread.LeadWindow, 'O̕ύX', False);
	Dialog.Owner := Garbage;

	Target := View.FocusedFile;

	if Target^.Kind = fkDrive then
		try
			SetString(Drive, Target^.Name, Target^.NameLength);
			Dialog.Value := GetVolumeLabel(Drive);
			Dialog.SelectAll;

			if Dialog.Execute then
			begin
				if not SetVolumeLabelW(Pointer(Drive), Pointer(PWideChar(Dialog.Value))) then
					RaiseLastOSError;

				Application.BeginWrite;
				try
					View.Model.Refresh
				finally
					Application.EndWrite;
				end
			end;
		except
		on E: EOSError do
			TMessageBox.Execute(E.Message, Name, MB_ICONERROR)
		end
	else
	begin
		BasePath := IncludeTrailingPathDelimiter(View.Model.Path.Directory);

		if View.SelectCount > 1 then
		begin
			Files := View.SelectedFiles;

			DisassembleFileName(Files[0].Name, NamePart, ExtPart);
			for I := High(Files) downto Succ(Low(Files)) do
			begin
				DisassembleFileName(Files[I].Name, N2, E2);
				if CompareFileName(NamePart, N2) <> 0 then NamePart := '*';
				if CompareFileName(ExtPart, E2) <> 0 then ExtPart := '*';
			end;

			if (NamePart = '*') and (ExtPart = '*') then
				TMessageBox.Execute('Oɋʕ܂', Name, MB_ICONWARNING)
			else
			begin
				RenameDialog := TRenameDialog.Create(CurrentThread.LeadWindow, 'O̕ύX');
				RenameDialog.Owner := Garbage;
				RenameDialog.FileName.Text := NamePart;
				RenameDialog.FileExt.Text := ExtPart;
				if RenameDialog.Execute then
				begin
					View.Model.StopSurveillance;
					try
						for I := High(Files) downto Low(Files) do
						begin
							DisassembleFileName(Files[I].Name, N2, E2);
							if NamePart <> '*' then N2 := RenameDialog.FileName.Text;
							if ExtPart <> '*' then E2 := RenameDialog.FileExt.Text;
							NewName := N2 + '.' + E2;
							if ShellCopyOrMoveOrRenameFile(CurrentThread.LeadWindow.Handle,
									BasePath + Files[I].Name, BasePath + NewName, FO_RENAME) then
							begin
								NotifyRename(BasePath + Files[I].Name, BasePath + NewName);
							end
							else
								TMessageBox.Execute('t@CύXł܂ł.', Name, MB_ICONERROR)
						end;
					finally
						View.Model.ResumeSurveillance(False);
					end;
				end
			end
		end
		else
		begin
			if View.SelectCount > 0 then
			begin
				for I := View.ItemCount - 1 downto 0 do
					if View.Selected[I] then Target := View.Items[I];
			end;

			SetString(OldName, Target^.Name, Target^.NameLength);

			if WideStrLScan(Pointer(OldName), Length(OldName), '\') <> nil then
			begin
				BasePath := ExtractFilePath(OldName);
				OldName := ExtractFileName(OldName);
			end;

			Dialog.Value := OldName;
			Dialog.CaretIndex := 0;
			Dialog.SelectedLength := Length(ChangeFileExt(OldName, ''));

			if Dialog.Execute then
			begin
				NewName := Dialog.Value;

				View.Model.StopSurveillance;
				R := False;
				try
					R := ShellCopyOrMoveOrRenameFile(CurrentThread.LeadWindow.Handle,
						BasePath + OldName, BasePath + NewName, FO_RENAME);
				finally
					View.Model.ResumeSurveillance(R);
				end;
				if R then
				begin
					View.FocusTo(NewName);
					NotifyRename(BasePath + OldName, BasePath + NewName);
				end
				else
					TMessageBox.Execute('t@CύXł܂ł.', Name, MB_ICONERROR)
			end
		end;
	end;
	Files := nil
end;

procedure TRenameAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	SetEnabled(View.SelectedIsRealFiles or
		(View.ItemCount > 0) and (View.FocusedFile.Kind = fkDrive));
end;

end.
