unit UActions;

{ EOFile by YT }

interface

uses
	Windows, Messages, ActiveX, ShellApi, ShlObj, CommCtrl, SysUtils, Types,
	SysUtilsW, SysHacks, AlgUtils, WinTips, ShlTips,
	LConsts, LObjects, LMenus, LWindows, LActions, LRegistry,
	LCmmDlgs, LAddDlgs, LStdCtls, LCmmCtls,
	EOFileIntf, UPath, UFolders, UView, UFileOp,
	UActionAttribute, UActionFindFile, UActionIncrementalSearch,
	UActionVersionInfo, UActionRename, UActionPack, UActionCommandLine,
	UActionCompareView, UActionDelete;

type
	TThreadActionManager = class(T0ActionManager)
	private
		procedure A_ClipboardCopy(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ClipboardCut(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ClipFileData(Action: TTransferAction; Order: TTransferOrder);
		procedure A_CloseView(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Console(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Copy(Action: TTransferAction; Order: TTransferOrder);
		procedure A_EnterFolder(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Explorer(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ExplorerFind(Action: TTransferAction; Order: TTransferOrder);
		procedure A_MakeDirectory(Action: TTransferAction; Order: TTransferOrder);
		procedure A_MakeShellLink(Action: TTransferAction; Order: TTransferOrder);
		procedure A_MarkAll(Action: TTransferAction; Order: TTransferOrder);
		procedure A_MarkReverse(Action: TTransferAction; Order: TTransferOrder);
		procedure A_MarkToggle(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Move(Action: TTransferAction; Order: TTransferOrder);
		procedure A_OpenFileAs(Action: TTransferAction; Order: TTransferOrder);
		procedure A_OpenFileShort(Action: TTransferAction; Order: TTransferOrder);
		procedure A_OpenView(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Property(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Sort(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToChosen(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToDesktop(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToFound(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToMyComputer(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToParent(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToRoot(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToShellLink(Action: TTransferAction; Order: TTransferOrder);
		procedure A_ToTemporary(Action: TTransferAction; Order: TTransferOrder);
	public
		procedure AfterConstruction; override;
	end;

	TClipboardPasteAction = class(T0Action)
	private
		FItems: TActionItems;
	protected
		function GetItems: IActionItems; override;
		procedure DoUpdateItems(AItems: TActionItems); override;
		procedure DoExecuteItem(AItem: TActionItem); override;
	public
		constructor Create;
		procedure Update; override;
	end;

	TOpenFileAction = class(T0Action)
	private
		FItems: TActionItems;
		FPopupMenu: TPopupMenu;
		FContextMenu: IContextMenu;
		FOpenAsFirst: DWORD;
		FOpenAsLast: DWORD;
		FUsing: Boolean;
	protected
		function GetItems: IActionItems; override;
		procedure DoExecuteItem(AItem: TActionItem); override;
		procedure DoUpdateItems(AItems: TActionItems); override;
		procedure DoExecute; override;
		function CustomDraw(TreeView: TTreeView; Node: TAnyTreeViewNode; Info: PNMTVCustomDraw): Boolean;
	public
		constructor Create;
		procedure Update; override;
	end;

	TDefaultAction = class(T0Action)
	private
		FTarget: IAction;
		FShort: Boolean;
	protected
		function GetItems: IActionItems; override;
		procedure DoExecute; override;
	public
		constructor Create(AShort: Boolean);
		procedure Update; override;
	end;

type
	TGlobalActionManager = class(T0ActionManager)
	private
		procedure A_BrowsePage(Action: TTransferAction; Order: TTransferOrder);
		procedure A_CloseWindow(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Dummy(Action: TTransferAction; Order: TTransferOrder);
		procedure A_HelpContents(Action: TTransferAction; Order: TTransferOrder);
		procedure A_HelpOnHelp(Action: TTransferAction; Order: TTransferOrder);
		procedure A_HelpTopicSearch(Action: TTransferAction; Order: TTransferOrder);
		procedure A_OpenWindow(Action: TTransferAction; Order: TTransferOrder);
		procedure A_Refresh(Action: TTransferAction; Order: TTransferOrder);
	protected
		function Select(const AItems: IActionItems): IActionItem; override;
	public
		procedure AfterConstruction; override;
	end;

	TFilterAction = class(T0Action)
	private
		FHistory: TWideStringDynArray;
	protected
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

	TMakeNewFileAction = class(T0Action)
	private
		FItems: TActionItems;
	protected
		function GetItems: IActionItems; override;
		procedure DoExecuteItem(AItem: TActionItem); override;
		procedure DoUpdateItems(AItems: TActionItems); override;
	public
		constructor Create;
		procedure Update; override;
	end;

	TWindowMenuAction = class(T0Action)
	private
		FItems: TActionItems;
	protected
		function GetItems: IActionItems; override;
		procedure DoExecuteItem(AItem: TActionItem); override;
		procedure DoUpdateItems(AItems: TActionItems); override;
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

	TExtensionMenuAction = class(T0Action)
	private
		FItems: TActionItems;
	protected
		function GetItems: IActionItems; override;
		procedure DoExecuteItem(AItem: TActionItem); override;
		procedure DoUpdateItems(AItems: TActionItems); override;
		procedure DoExecute; override;
	public
		constructor Create;
		procedure Update; override;
	end;

implementation

procedure OpenFileAs(const FileName: WideString);
var
	Hive: TRegistryHive;
	Reg: TRegistryKey;
	Cmd: WideString;
begin
	Hive := TRegistryHive.Create(HKEY_CLASSES_ROOT);
	Hive.Owner := Garbage;
	Reg := TRegistryKey.Create(Hive, 'Unknown\Shell\OpenAs\Command', KEY_READ);
	try
		Cmd := Reg.ReadString('');
	finally
		Reg.Free
	end;
	Cmd := WideStringReplace(Cmd, '%1', ExtractFileName(FileName),
					[rfReplaceAll, rfIgnorecase]);
	Cmd := ExpandEnvironmentStrings(Cmd);
	ExecuteCommandLine(Cmd, ExtractFileDir(FileName));
end;

{ TThreadActionManager }

procedure TThreadActionManager.AfterConstruction;
begin
	inherited;
	Add(A_ClipboardCopy);
	Add(A_ClipboardCut);
	Add(A_ClipFileData);
	Add(A_CloseView);
	Add(A_Console);
	Add(A_Copy);
	Add(A_EnterFolder);
	Add(A_Explorer);
	Add(A_ExplorerFind);
	Add(A_MakeDirectory);
	Add(A_MakeShellLink);
	Add(A_MarkAll);
	Add(A_MarkReverse);
	Add(A_MarkToggle);
	Add(A_Move);
	Add(A_OpenFileAs);
	Add(A_OpenFileShort);
	Add(A_OpenView);
	Add(A_Property);
	Add(A_Sort);
	Add(A_ToChosen);
	Add(A_ToDesktop);
	Add(A_ToFound);
	Add(A_ToMyComputer);
	Add(A_ToParent);
	Add(A_ToRoot);
	Add(A_ToShellLink);
  Add(A_ToTemporary);

	TCommandLineAction.Create.Owner := Self;
	TOpenFileAction.Create.Owner := Self;
	TAttributeAction.Create.Owner := Self;
	TIncrementalSearchAction.Create.Owner := Self;
	TClipboardPasteAction.Create.Owner := Self;
	TFindFileAction.Create.Owner := Self;
	TRenameAction.Create.Owner := Self;
	TUnpackAction.Create.Owner := Self;
	TPackAction.Create.Owner := Self;
	TDeleteAction.Create.Owner := Self;
	TCompareViewAction.Create.Owner := Self;

	TDefaultAction.Create(False).Owner := Self;
	TDefaultAction.Create(True).Owner := Self;
end;

procedure TThreadActionManager.A_ClipboardCopy;
var
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('ClipboardCopy', 'Nbv{[hփRs[',
						'It@CNbv{[hփRs[܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := Supports(View.Model, IDirectory) and (View.ItemCount > 0);
	end;
	toExecute:
	begin
		View := ActiveView;
		try
			OleSetClipboard(CreateFileListDataObject(View.Model.Path.Directory,
						FilePtrArray2WideFileNameArray(View.SelectedFiles),
						DROPEFFECT_COPY or DROPEFFECT_LINK, True));
			OleFlushClipboard;
		except
		on EInOutError do
			TMessageBox.Execute('VF猩Ȃt@Cł.', Action.Text, MB_ICONERROR);
		end;
	end;
	end;
end;

procedure TThreadActionManager.A_ClipboardCut;
var
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('ClipboardCut', 'Nbv{[h֐؂',
						'It@CNbv{[h֐؂܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := Supports(View.Model, IDirectory) and (View.ItemCount > 0);
	end;
	toExecute:
	begin
		View := ActiveView;
		OleSetClipboard(CreateFileListDataObject(View.Model.Path.Directory,
						FilePtrArray2WideFileNameArray(View.SelectedFiles),
						DROPEFFECT_MOVE));
		OleFlushClipboard;
	end;
	end;
end;

procedure TThreadActionManager.A_ClipFileData;
const
	PowerWitchFormat = 'PWtR Clipboard Format';
var
	View: TFilerView;
	Data: string;
	Mem: HGLOBAL;
	F: THANDLE;
	N: WideString;
	P: Pointer;
	CF_POWERWITCH: DWORD;
	Archive: IArchive;
begin
	case Order of
	toSetup:
		Action.Setup('ClipFileData', 't@Cf[^Rs[', 't@C̒gNbv{[hփRs[܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := (View.ItemCount > 0) and (View.SelectCount = 0) and
						(View.FocusedFile.Kind = fkFile) and
						(View.FocusedFile.Attributes and faDirectory = 0)
	end;
	toExecute:
		try
			View := ActiveView;
			if Supports(View.Model, IArchive, Archive) then
			begin
				Archive.UnpackMemory(View.FocusedFile, Data);
			end
			else
			begin
				with View.FocusedFile^ do SetString(N, Name, NameLength);
				F := FileOpen(ConcatPath(View.Model.Path.Directory, N), fmOpenRead or fmShareDenyWrite);
				if F = INVALID_HANDLE_VALUE then RaiseLastOSError;
				SetLength(Data, GetFileSize(F));
				FileRead(F, Pointer(Data)^, Length(Data));
				FileClose(F);
			end;
			{Rs[}
			if not OpenClipboard(CurrentThread.LeadWindow.Handle) then RaiseLastOSError;

			EmptyClipboard;

			Mem := GlobalAlloc(GMEM_MOVEABLE, Length(Data) + 1);
			P := GlobalLock(Mem);
			Move(PChar(Data)^, P^, Length(Data) + 1);
			GlobalUnlock(Mem);
			SetClipboardData(CF_TEXT, Mem);

			{PowerWitch}
			CF_POWERWITCH := RegisterClipboardFormatW(PowerWitchFormat);
			Mem := GlobalAlloc(GMEM_MOVEABLE, Length(Data) + 4);
			P := GlobalLock(Mem);
			PDWORD(P)^ := Length(Data);
			Move(Pointer(Data)^, Pointer(Cardinal(P) + 4)^, Length(Data));
			GlobalUnlock(Mem);
			SetClipboardData(CF_POWERWITCH, Mem);

			CloseClipboard;
		except
			TMessageBox.Execute('s܂', Action.Text, MB_ICONERROR);
		end;
	end;
end;

procedure TThreadActionManager.A_CloseView;
begin
	case Order of
	toSetup:
		Action.Setup('CloseView', 'r[', '݂̃r[܂');
	toUpdate:
		Action.Enabled := ActiveView.HostWindow.ViewCount > 1;
	toExecute:
		ActiveView.CloseOnDelay
	end;
end;

procedure TThreadActionManager.A_Console;
const
	ComSpec = 'COMSPEC';
var
	Cmd: array[0..1023] of WideChar;
begin
	case Order of
	toSetup:
		Action.Setup('Console', 'R\[', 'R\[N܂');
	toUpdate:
		Action.Enabled := Supports(ActiveView.Model, IDirectory);
	toExecute:
	begin
		GetEnvironmentVariableW(ComSpec, Cmd, Length(Cmd));
		ShellExecuteW(0, nil, Cmd, nil, PWideChar(ActiveView.Model.Path.Directory), SW_SHOW);
	end;
	end;
end;

procedure TThreadActionManager.A_Copy;
var
	ToPath: TPath;
	Manager: IFolderManager;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('Copy', 'Rs[...', 't@CRs[܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := View.SelectedIsRealFiles or
						Supports(View.Model, IArchive) and (View.ItemCount > 0)
	end;
	toExecute:
	begin
		if SelectFolder(Action.Text, ToPath) then
			try
				Manager := Application as IFolderManager;
				View := ActiveView;
				TFileOperationThread.Create(fokCopy, View.Model.Path, ToPath,
					FilePtrArray2WideFileNameArray(View.SelectedFiles));
			finally
				FinalizePath(ToPath);
			end;
	end;
	end;
end;

procedure TThreadActionManager.A_EnterFolder;
var
	View: TFilerView;
	SubFolder: IFolder;
	Item: PFile;
	P: TPath;
begin
	case Order of
	toSetup:
		Action.Setup('EnterFolder', 'tH_ɓ',
						'LbgtH_ɓ܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := (View.ItemCount > 0) and (foEnter in View.FocusedFile^.Operatable);
	end;
	toExecute:
	begin
		View := ActiveView;
		Item := View.FocusedFile;

		P := View.Model.ConcatenatePath(Item);
		SubFolder := (Application as IFolderManager).GetFolder(P);
		FinalizePath(P);

		if (SubFolder = nil) or not SubFolder.Valid then
			TMessageBox.Execute(P.Directory + 'ɓ܂.', Action.Text, MB_ICONERROR)
		else
			View.Model := SubFolder;
	end;
	end;
end;

procedure TThreadActionManager.A_Explorer;
begin
	case Order of
	toSetup:
		Action.Setup('Explorer', 'GNXv[', 'GNXv[N܂');
	toUpdate:
		Action.Enabled := Supports(ActiveView.Model, IDirectory);
	toExecute:
		ShellExecuteW(0, 'explore',
						PWideChar(ActiveView.Model.Path.Directory), nil, nil, SW_SHOW);
	end;
end;

procedure TThreadActionManager.A_ExplorerFind;
begin
	case Order of
	toSetup:
		Action.Setup('ExplorerFind', 't@CtH_',
						'GNXv[găt@CtH_܂');
	toUpdate:
		Action.Enabled := Supports(ActiveView.Model, IDirectory);
	toExecute:
		ShellExecuteW(0, 'find',
						PWideChar(ActiveView.Model.Path.Directory), nil, nil, SW_SHOW);
	end;
end;

procedure TThreadActionManager.A_MakeDirectory;
const
	Title = 'fBNg쐬';
var
	Dialog: TInputDialog;
	NewDir, Path, NewDirFullPath: TWideFileName;
	View: TFilerView;
	R: Boolean;
begin
	case Order of
	toSetup:
		Action.Setup('MakeDirectory', 'fBNg쐬...',
						'fBNgVKɍ쐬܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := fpPlaceFile in View.Model.Possibilities
	end;
	toExecute:
	begin
		View := ActiveView;

		Dialog := TInputDialog.Create(CurrentThread.LeadWindow, Title, False);
		Dialog.Owner := Garbage;
		Dialog.Value := 'Untitled';
		if Dialog.Execute then
		begin
			NewDir := Dialog.Value;
			Path := View.Model.Path.Directory;
			NewDirFullPath := ConcatPath(Path, NewDir);

			View.Model.StopSurveillance;
			R := CreateDirectoryW(PWideChar(NewDirFullPath), nil);
			View.Model.ResumeSurveillance(True);

			if R then
			begin
				if Supports(View.Model, IDirectory) then
					View.FocusTo(NewDir)
				else
					View.FocusTo(NewDirFullPath);
			end
			else
				TMessageBox.Execute(NewDir + '͖ȖOł', Title, MB_ICONERROR)
		end;
	end;
	end;
end;

procedure TThreadActionManager.A_MakeShellLink;
var
	ToPath: TPath;
	Manager: IFolderManager;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('MakeShellLink', 'V[gJbg쐬...',
						'V[gJbg쐬܂');
	toUpdate:
		Action.Enabled := ActiveView.SelectedIsRealFiles;
	toExecute:
	begin
		if SelectFolder(Action.Text, ToPath) then
			try
				Manager := Application as IFolderManager;
				View := ActiveView;
				TFileOperationThread.Create(fokLink, View.Model.Path, ToPath,
					FilePtrArray2WideFileNameArray(View.SelectedFiles));
			finally
				FinalizePath(ToPath);
			end;
	end;
	end;
end;

procedure TThreadActionManager.A_MarkAll;
var
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('MarkAll', 'ׂđI', 'SĂ̍ڂI܂');
	toExecute:
	begin
		View := ActiveView;
		View.BeginUpdate;
		View.SelectAll(View.SelectCount = 0);
		View.EndUpdate;
	end;
	end;
end;

procedure TThreadActionManager.A_MarkReverse;
var
	View: TFilerView;
	I: Integer;
begin
	case Order of
	toSetup:
		Action.Setup('MarkReverse', 'I𔽓]', 'ڂ̑I𔽓]܂');
	toExecute:
	begin
		View := ActiveView;
		for I := View.ItemCount - 1 downto 0 do
			View.Selected[I] := not View.Selected[I];
	end;
	end;
end;

procedure TThreadActionManager.A_MarkToggle;
var
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('MarkToggle', 'I/I',	'ڂI/I܂');
	toExecute:
	begin
		View := ActiveView;
		View.Selected[View.CaretIndex] := not View.Selected[View.CaretIndex];
		if View.CaretIndex < View.ItemCount - 1 then
			View.CaretIndex := View.CaretIndex + 1;
	end;
	end;
end;

procedure TThreadActionManager.A_Move;
var
	ToPath: TPath;
	Manager: IFolderManager;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('Move', 'ړ...', 't@Cړ܂');
	toUpdate:
		Action.Enabled := ActiveView.SelectedIsRealFiles;
	toExecute:
	begin
		if SelectFolder(Action.Text, ToPath) then
			try
				Manager := Application as IFolderManager;
				View := ActiveView;
				TFileOperationThread.Create(fokMove, View.Model.Path, ToPath,
					FilePtrArray2WideFileNameArray(View.SelectedFiles));
			finally
				FinalizePath(ToPath);
			end;
	end;
	end;
end;

procedure TThreadActionManager.A_OpenFileAs;
var
	TargetFile: PFile;
	FullPathName: TWideFileName;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('OpenFileAs', 'AvP[VJ...',
						'AvP[VIŃt@CJ܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := (View.ItemCount > 0) and (View.SelectCount = 0) and
						(View.FocusedFile.Kind = fkFile) and
						(View.FocusedFile.Attributes and faDirectory = 0) and
						not Supports(View.Model, IArchive);
	end;
	toExecute:
	begin
		View := ActiveView;
		TargetFile := View.FocusedFile;
		SetString(FullPathName, TargetFile.Name, TargetFile.NameLength);
		if ExtractFileDir(FullPathName) = '' then
			FullPathName := IncludeTrailingPathDelimiter(View.Model.Path.Directory) + FullPathName;
		OpenFileAs(FullPathName);
	end;
	end;
end;

procedure TThreadActionManager.A_OpenFileShort;
var
	TargetFile: PFile;
	FullPathName: TWideFileName;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('OpenFileShort', '₢킹ɊJ', 't@CftHgŊJ܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := (View.ItemCount > 0) and (View.SelectCount = 0) and
						(View.FocusedFile.Kind = fkFile) and
						(View.FocusedFile.Attributes and faDirectory = 0) and
						not Supports(View.Model, IArchive);
	end;
	toExecute:
	begin
		View := ActiveView;
		TargetFile := View.FocusedFile;
		SetString(FullPathName, TargetFile.Name, TargetFile.NameLength);
		if ExtractFileDir(FullPathName) = '' then
			FullPathName := IncludeTrailingPathDelimiter(View.Model.Path.Directory) + FullPathName;
		if ShellExecuteW(0, nil, PWideChar(FullPathName), nil, PWideChar(View.Model.Path.Directory), SW_SHOW) <= 32 then
			OpenFileAs(FullPathName);
	end;
	end;
end;

procedure TThreadActionManager.A_OpenView;
var
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('OpenView', 'ʂ̃r[J', 'r[ЂƂJ܂');
	toExecute:
	begin
		View := ActiveView;
		UView.OpenView(View, View.Model.Path);
	end;
	end;
end;

procedure TThreadActionManager.A_Property;
var
	NameSpace: INameSpace;
	View: TFilerView;
	Target: TPath;
	TargetFile: WideString;
begin
	case Order of
	toSetup:
		Action.Setup('Property', 'vpeB', 'vpeBV[g\');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := (View.ItemCount > 0) and (View.SelectCount = 0);
		if Action.Enabled then
			Action.Enabled := foProperty in View.FocusedFile.Operatable;
	end;
	toExecute:
	begin
		View := ActiveView;
		if Supports(View.Model, INameSpace, NameSpace) then
		begin
			Target := NameSpace.ConcatenatePath(View.FocusedFile);
			if Target.ItemIDList = nil then
				ShowPropertySheet(View.HostWindow.Handle, Target.Directory)
			else
				ShowPropertySheet(View.HostWindow.Handle, Target.ItemIDList);
			FinalizePath(Target);
		end
		else
		begin
			TargetFile := ConcatPath(View.Model.Path.Directory, View.FocusedFile.Name);
			ShowPropertySheet(View.HostWindow.Handle, TargetFile);
		end;
	end;
	end;
end;

procedure TThreadActionManager.A_Sort;

	const S: array[TSortMode] of WideString = (
		't@CVXe',
		'O()', '',
		'O', '',
		'gq', '',
		'TCY', '',
		'ŏIXV', '',
		'ŏIQƓ', '');

	var D: TSelectDialog;
	var CheckBox: TCheckBox;

	procedure SelectionChange; cdecl;
	var
		R: Boolean;
	begin
		R := D.List.SelectedIndex > 0;
		CheckBox.Enabled := R;
		if not R then CheckBox.Checked := False;
	end;

var
	I, Se: TSortMode;
	E: T0ControlItem;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('Sort', 'т...', '\т܂');
	toUpdate:
		Action.Enabled := fpSort in ActiveView.Model.Possibilities;
	toExecute:
	begin
		View := ActiveView;
		D := TSelectDialog.Create(CurrentThread.LeadWindow, Action.Text);
		D.Owner := Garbage;
		D.Context := GetContext;
		D.OnSelectionChange := Pointer(@SelectionChange);
		CheckBox := TCheckBox.Create(D, 't(&R)', ZeroPoint, DefaultSize);
		E := nil;
		for I := Low(TSortMode) to High(TSortMode) do
		begin
			if S[I] <> '' then
			begin
				E := D.Add;
				E.Text := S[I];
				E.Data := Pointer(I);
				if View.SortMode = I then D.Selected := E;
			end
			else
			begin
				if View.SortMode = I then
				begin
					D.Selected := E;
					CheckBox.Checked := True;
				end
			end
		end;
		D.SpaceSize := CheckBox.BoundsSize;
		CheckBox.Position := D.SpaceRect.TopLeft;
		SelectionChange;
		if D.Execute then
		begin
			Se := TSortMode(D.Selected.Data);
			if CheckBox.Checked then Inc(Se);
			View.SortMode := Se
		end;
	end;
	end;
end;

procedure TThreadActionManager.A_ToChosen;
var
	Dialog: TBrowseFolderDialog;
	P: TPath;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('ToChosen', 'Iňړ', 'tH_uEYĈړ');
	toExecute:
	begin
		View := ActiveView;

		Dialog := TBrowseFolderDialog.Create(CurrentThread.LeadWindow, Action.Text);
		Dialog.Owner := Garbage;
		Dialog.Styles := [bfShowText];
		Dialog.AutoShowPath := True;
		Dialog.Directory := View.Model.Path.Directory;

		if Dialog.Execute then
		begin
			InitializePath(P);

			if (Dialog.Directory <> '') and (Dialog.ItemIDList.mkid.cb <> 0) then
				P.Directory := Dialog.Directory
			else
				P.ItemIDList := Dialog.ItemIDList;

			View.Model := (Application as IFolderManager).GetFolder(P);
		end;
	end;
	end;
end;

procedure TThreadActionManager.A_ToDesktop;
const
	Desktop: ItemIDList = (mkid: (cb: 0));
	DesktopPath: TPath = (ItemIDList: @Desktop);
var
	NameSpace: INameSpace;
begin
	case Order of
	toSetup:
		Action.Setup('ToDesktop', 'fXNgbvɈړ', 'fXNgbvɈړ܂');
	toUpdate:
		Action.Enabled := not Supports(ActiveView.Model, INameSpace, NameSpace)
						or not NameSpace.AtDesktop;
	toExecute:
	begin
		ActiveView.Model := (Application as IFolderManager).GetFolder(DesktopPath);
	end;
	end;
end;

procedure TThreadActionManager.A_ToFound;
var
	View: TFilerView;
	P: TPath;
	S: WideString;
begin
	case Order of
	toSetup:
		Action.Setup('ToFound', '݈ʒuֈړ', 'ʂ̃t@C݈̑ʒuɈړ܂');
	toUpdate:
		with ActiveView do
			Action.Enabled := Supports(Model, IFounds) and (ItemCount > 0);
	toExecute:
	begin
		View := ActiveView;
		with View.FocusedFile^ do SetString(S, Name, NameLength);
		InitializePath(P);
		P.Directory := ExtractFileDir(S);
		View.Model := (Application as IFolderManager).GetFolder(P);
		View.FocusTo(ExtractFileName(S))
	end;
	end;
end;

procedure TThreadActionManager.A_ToMyComputer;
var
	P: TPath;
begin
	case Order of
	toSetup:
		Action.Setup('ToMyComputer', '}CRs[^Ɉړ',
						'}CRs[^Ɉړ܂');
	toUpdate:
		Action.Enabled := not IsDrives(ActiveView.Model);
	toExecute:
	begin
		InitializePath(P);
		SHGetSpecialFolderLocation(0, CSIDL_DRIVES, P.ItemIDList);
		try
			ActiveView.Model := (Application as IFolderManager).GetFolder(P);
		finally
			FinalizePath(P);
		end
	end;
	end;
end;

procedure TThreadActionManager.A_ToParent;
var
	View: TFilerView;
	CurrentName: TWideFileName;
	P: PWideChar;
begin
	case Order of
	toSetup:
		Action.Setup('ToParent', '', 'ЂƂ̃tH_Ɉړ܂');
	toUpdate:
		Action.Enabled := fpHasParent in ActiveView.Model.Possibilities;
	toExecute:
	begin
		View := ActiveView;
		Application.BeginWrite;
		try
			CurrentName := '';
			if Supports(View.Model, IDirectory) then
			begin
				CurrentName := ExtractFileName(View.Model.Path.Directory);
				if CurrentName = '' then
				begin
					CurrentName := ExtractFileDrive(View.Model.Path.Directory);
					if CurrentName[2] = ':' then CurrentName := CurrentName + '\'
				end;
			end
			else if Supports(View.Model, INameSpace) then
			begin
				CurrentName := View.Model.Path.Directory;
				P := PWideChar(CurrentName);
				CurrentName := Copy(CurrentName, WideStrRScan(P, ':') - P, MaxInt);
			end
			else if Supports(View.Model, IArchive) then
			begin
				CurrentName := View.Model.Path.Archive;
				if CurrentName = '' then
					CurrentName := ExtractFileName(View.Model.Path.Directory)
				else
				begin
					P := WideStrRScan(PWideChar(CurrentName), '/');
					if P <> nil then Delete(CurrentName, 1, P - Pointer(CurrentName) + 1);
				end;
			end;
			View.Model := GetParentFolder(View.Model);
			View.FocusTo(CurrentName);
		finally
			Application.EndWrite;
		end
	end;
	end;
end;

procedure TThreadActionManager.A_ToRoot;
var
	Path: TPath;
	//S: Integer;
	View: TFilerView;
begin
	case Order of
	toSetup:
		Action.Setup('ToRoot', '[gfBNgɈړ',
						'[gfBNgɈړ܂');
	toUpdate:
	begin
		View := ActiveView;
		Action.Enabled := not IsRootDir(View.Model);
	end;
	toExecute:
	begin
		View := ActiveView;
		InitializePath(Path);
		Path.Directory := ExtractFileDrive(View.Model.Path.Directory);
		if (Length(Path.Directory) = 2) and (Path.Directory[2] = ':') then
			Path.Directory := Path.Directory + '\';

		//S := Application.BeginUpdate;
		//try
			View.Model := (Application as IFolderManager).GetFolder(Path);
		//finally
		//	Application.EndUpdate(S)
		//end;
	end;
	end;
end;

procedure TThreadActionManager.A_ToShellLink;
var
	View: TFilerView;
	Name: TWideFileName;
	//S: Integer;
	Manager: IFolderManager;
	Path: TPath;
	Focused: PFile;
	Folder: IFolder;
begin
	case Order of
	toSetup:
		Action.Setup('ToShellLink', 'Nֈړ', 'V[gJbgֈړ܂');
	toUpdate:
	begin
		View := ActiveView;
		if View.CaretIndex < 0 then
			Action.Enabled := False
		else
		begin
			Focused := View.FocusedFile;
			SetString(Name, Focused.Name, Focused.NameLength);
			Action.Enabled := WideSameText(ExtractFileExt(Name), '.lnk');
		end
	end;
	toExecute:
	try
		Manager := Application as IFolderManager;
		View := ActiveView;
		Focused := View.FocusedFile;
		SetString(Name, Focused.Name, Focused.NameLength);
		Name := ConcatPath(View.Model.Path.Directory, Name);

		Path := Manager.LinkToPath(Name);
		try
			//S := Application.BeginUpdate;
			//try
				Folder := Manager.GetFolder(Path);
				if Folder.Valid then
					View.Model := Folder
				else
				begin
					View.Model := GetParentFolder(Folder);
					if Supports(Folder, INameSpace) then
						View.FocusTo(Folder.DisplayText)
					else
						View.FocusTo(ExtractFileName(Path.Directory));
				end;
			//finally
			//	Application.EndUpdate(S)
			//end
		finally
			FinalizePath(Path);
		end
	except
	on E: Exception do
		TMessageBox.Execute('V[gJbg̉Ɏs܂.'#13#10
			+ '(񍐎͂LĂˁ ' + E.ClassName + ' : ' + E.Message + ')',
			Action.Text, MB_ICONERROR);
	end;
	end
end;

procedure TThreadActionManager.A_ToTemporary;
var
	Path: TPath;
	Model: IFolder;
begin
	case Order of
	toSetup:
		Action.Setup('ToTemporary', 'e|Ɉړ', 'e|fBNgɈړ܂');
	toUpdate:
	begin
		Model := ActiveView.Model;
		Action.Enabled := not Supports(Model, IDirectory) or
			(Model.Path.Directory <> ExtractLongPathName(TemporaryDirectory));
	end;
	toExecute:
	begin
		InitializePath(Path);
		Path.Directory := ExtractLongPathName(TemporaryDirectory);
		ActiveView.Model := (Application as IFolderManager).GetFolder(Path);
	end;
	end;
end;

{ TClipboardPasteAction }

constructor TClipboardPasteAction.Create;
begin
	inherited Create(False);
	SetName('ClipboardPaste');
	SetText('Nbv{[h\t');
	SetHint('Nbv{[hɋL^ꂽt@C\t܂');

	FItems := TActionItems.Create(Self);
	FItems.Owner := Self;
	FItems.Title := 'Nbv{[hɋL^ꂽt@C'
end;

procedure TClipboardPasteAction.DoExecuteItem(AItem: TActionItem);
var
	D: IDataObject;
begin
	if OleGetClipboard(D) = S_OK then
		ActiveView.AcceptDataObject(D, DWORD(AItem.Data));
end;

procedure TClipboardPasteAction.DoUpdateItems(AItems: TActionItems);
const
	CF_DROPEFFECT = 'Preferred DropEffect';
	Force: array[Boolean] of String = (
					'ʂL^ĂȂӒnƓxŋI', '');
	Modes: array[0..2] of record M: DWORD; T: String end = (
					(M: DROPEFFECT_COPY; T: 'Rs['),
					(M: DROPEFFECT_MOVE; T: 'ړ'),
					(M: DROPEFFECT_LINK; T: 'N'));
var
	D: IDataObject;
	Effect: DWORD;
	EffectFormat: FORMATETC;
	M: STGMEDIUM;
	P: Pointer;
	I: Boolean;
	J: Integer;
begin
	AItems.Clear;

	if OleGetClipboard(D) = S_OK then
	begin
		EffectFormat := HDropFormat;
		EffectFormat.cfFormat := RegisterClipboardFormatW(CF_DROPEFFECT);
		if D.GetData(EffectFormat, M) = S_OK then
		begin
			P := GlobalLock(M.hGlobal);
			Effect := DWORD(P^);
			GlobalUnlock(M.hGlobal);
			ReleaseStgMedium(M);
		end
		else
			Effect := DROPEFFECT_MOVE;

		for I := True downto False do
			for J := Low(Modes) to High(Modes) do
				if (Effect and Modes[J].M <> 0) = I then
					with TActionItem.Create(AItems, nil) do
					begin
						Data := Pointer(Modes[J].M);
						Text := Force[I] + Modes[J].T;
					end;
	end;
end;

function TClipboardPasteAction.GetItems: IActionItems;
begin
	Result := FItems
end;

procedure TClipboardPasteAction.Update;
var
	D: IDataObject;
	E: Boolean;
begin
	E := False;

	if Supports(ActiveView.Model, IDirectory) then
		if OleGetClipboard(D) = S_OK then
			if D.QueryGetData(HDropFormat) = S_OK then
				E := True;

	SetEnabled(E);
	FItems.PostUpdate
end;

{ TOpenFileAction }

constructor TOpenFileAction.Create;
begin
	inherited Create(False);

	FItems := TActionItems.Create(Self);
	FItems.Owner := Self;

	SetName('OpenFile');
	SetText('J');
	SetHint('t@CJ܂');
end;

function TOpenFileAction.CustomDraw(TreeView: TTreeView; Node: TAnyTreeViewNode; Info: PNMTVCustomDraw): Boolean;
var
	ID: DWORD; {Word}
	CM2: IContextMenu2;
	DrawItemStruct: TDrawItemStruct;
	MeasureItemStruct: TMeasureItemStruct;
	ItemData: DWORD;
begin
	Result := False;
	if (Win32MajorVersion = 5) and (Win32MinorVersion = 1) and
		(Node <> nil) and (Node.Level > 0) and (Node.Data <> nil) then
	begin
		ID := DWORD((IActionItem(Node.Data) as IActionItemData).Data);
		if (ID <> 0) and (ID > FOpenAsFirst) and (ID <= FOpenAsLast) and
			(Info^.nmcd.rc.Top < Info^.nmcd.rc.Bottom) and
			Supports(FContextMenu, IContextMenu2, CM2) then
		begin
			ItemData := ID - FOpenAsFirst - 1;

			MeasureItemStruct.CtlType := ODT_MENU;
			MeasureItemStruct.CtlID := 0;
			MeasureItemStruct.itemID := ID;
			MeasureItemStruct.itemWidth := 0;
			MeasureItemStruct.itemHeight := 0;
			MeasureItemStruct.itemData := ItemData;
			CM2.HandleMenuMsg(WM_MEASUREITEM, 0, LPARAM(@MeasureItemStruct));

			DrawItemStruct.CtlType := ODT_MENU;
			DrawItemStruct.CtlID := 0;
			DrawItemStruct.itemID := ID;
			DrawItemStruct.itemAction := ODA_SELECT;
			DrawItemStruct.itemState := ODS_DEFAULT;
			if Info^.nmcd.uItemState and CDIS_SELECTED <> 0 then DrawItemStruct.itemState := ODS_SELECTED;
			DrawItemStruct.hwndItem := 0;
			DrawItemStruct.hDC := Info^.nmcd.hdc;
			DrawItemStruct.rcItem := Info^.nmcd.rc;
			Inc(DrawItemStruct.rcItem.Left, Node.Level * TreeView.Indent);
			Dec(DrawItemStruct.rcItem.Left, GetTextSize(DrawItemStruct.hDC, #32).cx);
			DrawItemStruct.itemData := ItemData;

			SaveDC(DrawItemStruct.hDC);
			CM2.HandleMenuMsg(WM_DRAWITEM, 0, LPARAM(@DrawItemStruct));
			RestoreDC(DrawItemStruct.hDC, -1);
			Result := True;
		end;
	end;
end;

procedure TOpenFileAction.DoExecute;
var
	TheItems: IActionItems;
	Selected: IActionItem;
begin
	FUsing := True;
	try
		TheItems := GetItems;
		if TheItems <> nil then
		begin
			Selected := SelectActionItemsDialog(TheItems, CustomDraw);
			if Selected <> nil then
				Selected.Execute
		end
	finally
		FUsing := False
	end;
end;

procedure TOpenFileAction.DoExecuteItem(AItem: TActionItem);
var
	CmdInfo: CMINVOKECOMMANDINFOEX;
	View: TFilerView;
	WideDir: WideString;
	AnsiDir: AnsiString;
	ID: DWORD;
	//Cmd: array[0..255] of Char;
begin
	View := ActiveView;

	WideDir := View.Model.Path.Directory;
	AnsiDir := WideDir;

	//Cmd[0] := #0;
	//FContextMenu.GetCommandString(Cardinal(AItem.Data), GCS_VERBA, nil, Cmd, 255);
	//OutputDebugStringA(Cmd);

	ID := DWORD(AItem.Data) - $1000;

	FillChar(CmdInfo, SizeOf(CmdInfo), 0);
	CmdInfo.cbSize  := SizeOf(CmdInfo);
	CmdInfo.fMask   := CMIC_MASK_UNICODE;
	CmdInfo.HWnd    := View.Handle;
	CmdInfo.nShow   := SW_SHOWNORMAL;
	{Ansi}
	CmdInfo.lpVerb      := PAnsiChar(ID);
	CmdInfo.lpDirectory := PChar(AnsiDir);
	{Wide}
	CmdInfo.lpVerbW      := PWideChar(ID);
	CmdInfo.lpDirectoryW := PWideChar(WideDir);

	FContextMenu.InvokeCommand(PCMINVOKECOMMANDINFO(@CmdInfo)^);
end;

procedure TOpenFileAction.DoUpdateItems(AItems: TActionItems);
var
	CM2: IContextMenu2;

	procedure Enum(const Items: TActionItems; Menu: TAnyMenu);
	var
		O: Integer;

		function Each(MenuItem: TAnyMenuItem): Boolean; cdecl;
		var
			S: TAnyMenu;
			ActionItem: TActionItem;
		begin
			if not MenuItem.Separator then
			begin
				if Menu <> FPopupMenu then
				begin
					if MenuItem.ID < FOpenAsFirst then FOpenAsFirst := MenuItem.ID;
					if MenuItem.ID > FOpenAsLast then FOpenAsLast := MenuItem.ID;
				end;

				ActionItem := TActionItem.Create(Items, nil);
				ActionItem.Text := MenuItem.Text;
				ActionItem.Data := Pointer(MenuItem.ID);

				S := MenuItem.Sub;
				if S <> nil then
				begin
					if CM2 <> nil then
						CM2.HandleMenuMsg(WM_INITMENUPOPUP, S.Handle, O);

					Enum(ActionItem.ReadyItems, S);
				end
			end;
			Inc(O);

			Result := False;
		end;

	begin
		O := 0;
		Menu.EnumerateItems(CEachMenuItem(@Each), GetContext);
	end;

var
	F: TWideFileName;
	View: TFilerView;
{$IFOPT C+}
	R: Integer;
{$ENDIF}
begin
	if AItems = FItems then
	begin
		AItems.Clear;
		FOpenAsFirst := High(FOpenAsFirst);
		FOpenAsLast := 0;

		View := ActiveView;
{$IFOPT C+}
		R := TOpenFileAction(View)._RefCount;
{$ENDIF}
		with View.FocusedFile^ do SetString(F, Name, NameLength);
		AItems.Title := F + 'J';

		FContextMenu := View.QueryContextMenu;
		if FContextMenu <> nil then
		begin
			//|bvAbvj[͒Kv炵
			FPopupMenu.Free;
			FPopupMenu := TPopupMenu.Create;
			FPopupMenu.Owner := Self;

			FContextMenu.QueryContextMenu(FPopupMenu.Handle, 0, $1000, $7fff, CMF_VERBSONLY);

			if Supports(FContextMenu, IContextMenu2, CM2) then
				CM2.HandleMenuMsg(WM_INITMENUPOPUP, FPopupMenu.Handle, 0);

			Enum(FItems, FPopupMenu);
		end;

{$IFOPT C+}
		if R <> TOpenFileAction(View)._RefCount then
			TMessageBox.Execute('RefCount Changed! @TOpenFileAction.DoUpdateItems', MB_ICONWARNING);
{$ENDIF}
	end;
end;

function TOpenFileAction.GetItems: IActionItems;
begin
	Result := FItems
end;

procedure TOpenFileAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	if not (opInMenuLoop in View.HostWindow.Operating) and not FUsing then
	begin
		FContextMenu := nil;
		FPopupMenu.Free;
		FPopupMenu := nil;

		SetEnabled((View.ItemCount > 0) and (View.SelectCount = 0) and
						(View.FocusedFile.Kind = fkFile) and
						(View.FocusedFile.Attributes and faDirectory = 0) and
						not Supports(View.Model, IArchive));

		FItems.PostUpdate
	end;
end;

{ TDefaultAction }

constructor TDefaultAction.Create(AShort: Boolean);
const
	S: array[Boolean] of string = ('Default', 'DefaultShort');
begin
	FShort := AShort;
	inherited Create(False);
	SetName(S[AShort]);
end;

procedure TDefaultAction.DoExecute;
begin
	FTarget.Execute
end;

function TDefaultAction.GetItems: IActionItems;
begin
	Result := FTarget.Items
end;

procedure TDefaultAction.Update;
	function UpdateBy(const AActionName: String): Boolean;
	var
		A: IAction;
	begin
		Assert(Owner is TThreadActionManager);
		A := TThreadActionManager(Owner).Actions[AActionName];
		Result := A.Enabled;
		if Result then
		begin
			SetText(A.Text);
			SetHint(A.Hint);
			SetEnabled(True);
			FTarget := A;
		end;
	end;
begin
	if UpdateBy('EnterFolder') then Exit;
	if FShort and UpdateBy('OpenFileShort') then Exit;
	if UpdateBy('ToShellLink') then Exit;
	if UpdateBy('OpenFile') then Exit;
	if UpdateBy('OpenFileAs') then Exit;
	UpdateBy('Dummy'); SetEnabled(False)
end;

{ TGlobalActionManager }

procedure TGlobalActionManager.AfterConstruction;
begin
	inherited;
	Add(A_BrowsePage);
	Add(A_CloseWindow);
	Add(A_Dummy);
	Add(A_HelpContents);
	Add(A_HelpOnHelp);
	Add(A_HelpTopicSearch);
	Add(A_OpenWindow);
	Add(A_Refresh);

	TFilterAction.Create.Owner := Self;
	TMakeNewFileAction.Create.Owner := Self;
	TWindowMenuAction.Create.Owner := Self;
	TVersionInfoAction.Create.Owner := Self;
  TExtensionMenuAction.Create.Owner := Self;
end;

procedure TGlobalActionManager.A_BrowsePage;
const
	URL = 'http://panathenaia.halfmoon.jp/eofile/';
begin
	case Order of
	toSetup:
		Action.Setup('BrowsePage', 'EOFilẽy[W\',
						'EOFilẽy[WuEUŊJ܂');
	toExecute:
		ShellExecuteW(0, nil, URL, nil, nil, SW_SHOW);
	end;
end;

procedure TGlobalActionManager.A_CloseWindow;
begin
	case Order of
	toSetup:
		Action.Setup('CloseWindow', '', '̑܂');
	toExecute:
		ActiveView.HostWindow.Close
	end;
end;

procedure TGlobalActionManager.A_Dummy;
begin
	case Order of
	toSetup:
	begin
		Action.Setup('Dummy', SNothing, '܂');
		Action.Enabled := False;
	end;
	end;
end;

procedure TGlobalActionManager.A_HelpContents;
begin
	case Order of
	toSetup:
		Action.Setup('HelpContents', 'wv', 'wv̖ڎ\܂');
	toExecute:
		WinHelpW(T0GUIThread(CurrentThread).MainWindow.Handle,
						PWideChar(Application.ExePath + 'EOFile.hlp'), HELP_CONTENTS, 0);
	end;
end;

procedure TGlobalActionManager.A_HelpOnHelp;
begin
	case Order of
	toSetup:
		Action.Setup('HelpOnHelp', 'wv̎g',
						'Windows ̃wv̎g\܂');
	toExecute:
		WinHelpW(T0GUIThread(CurrentThread).MainWindow.Handle,
						'', HELP_HELPONHELP, LPARAM(PChar('')));

	end;
end;

procedure TGlobalActionManager.A_HelpTopicSearch;
begin
	case Order of
	toSetup:
		Action.Setup('HelpTopicSearch', 'gsbŇ',
						'wṽgsbN܂');
	toExecute:
		WinHelpW(T0GUIThread(CurrentThread).MainWindow.Handle,
						PWideChar(Application.ExePath + 'EOFile.hlp'), HELP_PARTIALKEY,
						LPARAM(PChar('')));
	end;
end;

procedure TGlobalActionManager.A_OpenWindow;
begin
	case Order of
	toSetup:
		Action.Setup('OpenWindow', 'ʂ̑J', 'ʂ̑J܂');
	toExecute:
		(Application as IEOFileGlobal).StartViewThread(ActiveView.Model.Path)
	end;
end;

procedure TGlobalActionManager.A_Refresh;
begin
	case Order of
	toSetup:
		Action.Setup('Refresh', 'XV', 'ŐV̏ɍXV܂');
	toExecute:
	begin
		Application.BeginWrite;
		try
			ActiveView.Model.Refresh
		finally
			Application.EndWrite;
		end;
	end;
	end;
end;

function TGlobalActionManager.Select(const AItems: IActionItems): IActionItem;
begin
	Result := SelectActionItemsDialog(AItems)
end;

{ TFilterAction }

constructor TFilterAction.Create;
begin
	inherited Create(True);

	SetName('Filter');
	SetText('tB^...');
	SetHint('ChJ[hŃtB^܂');

	SetLength(FHistory, 2);
	FHistory[0] := '"*.*"';
	FHistory[1] := 'not Hidden';
end;

procedure TFilterAction.DoExecute;
var
	View: TFilerView;
	Dialog: TInputDialog;
	I: Integer;
	W: WideString;
begin
	View := ActiveView; {eXbh}

	Dialog := TInputDialog.Create(CurrentThread.LeadWindow, Text, True);
	Dialog.Owner := Garbage;

	for I := Low(FHistory) to High(FHistory) do Dialog.Add.Text := FHistory[I];

	if View.WildCard = '' then
	begin
		Dialog.Value := '"*.*"';
		Dialog.CaretIndex := 1;
		Dialog.SelectedLength := 3;
	end
	else
		Dialog.Value := View.WildCard;

	if Dialog.Execute then
		try
			W := Dialog.Value;
			View.WildCard := W;

			Application.BeginWrite;
			try
				AddHistory(FHistory, W, MaxInt);
			finally
				Application.EndWrite;
			end;
		except
		on E: Exception do
			TMessageBox.Execute(E.Message, Text, MB_ICONERROR);
		end;
end;

procedure TFilterAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	SetEnabled((View <> nil) and Supports(View.Model, IDirectory));
end;

{ TMakeNewFileAction }

type
	TMakeNewFileKind = (nfkNullFile, nfkFileName);
	TMakeNewFileActionItem = class(TActionItem)
	public
		FileExt: TFileName;
		Kind: TMakeNewFileKind;
		FileName: TFileName;
	end;

constructor TMakeNewFileAction.Create;
begin
	inherited Create(True);

	FItems := TActionItems.Create(Self);
	FItems.Owner := Self;

	SetName('MakeNewFile');
	SetText('t@CVK쐬');
	SetHint('Vt@C쐬܂');
end;

procedure TMakeNewFileAction.DoExecuteItem(AItem: TActionItem);
const
	Untitled = 'Untitled';
var
	Dialog: TInputDialog;
	NewName, NewFileName, SourceFileName: TWideFileName;
	View: TFilerView;
	Path: TPath;
begin
	Assert(AItem is TMakeNewFileActionItem);

	View := ActiveView;
	Path := View.Model.Path; {MȂ̂Ŕjsv}

	Dialog := TInputDialog.Create(CurrentThread.LeadWindow, Text + ' ' + AItem.Text, False);
	Dialog.Owner := Garbage;

	Dialog.Value := Untitled + TMakeNewFileActionItem(AItem).FileExt;
	Dialog.CaretIndex := 0;
	Dialog.SelectedLength := Length(Untitled);
	if Dialog.Execute then
	begin
		NewName := Dialog.Value;
		NewFileName := IncludeTrailingPathDelimiter(Path.Directory) + NewName;

		if FileSystemExists(NewFileName) then
		begin
			TMessageBox.Execute(NewFileName + '͊ɂ܂', Text, MB_ICONERROR);
			Exit;
		end;

		View.Model.StopSurveillance;
		try
			case TMakeNewFileActionItem(AItem).Kind of
			nfkFileName:
			begin
				SourceFileName := SpecialFolderPath(0, CSIDL_TEMPLATES) + TMakeNewFileActionItem(AItem).FileName;
				if not FileExists(SourceFileName) then
				begin
					SourceFileName := SpecialFolderPath(0, 45) + TMakeNewFileActionItem(AItem).FileName;
					if not FileExists(SourceFileName) then
					begin
						SourceFileName := WindowsPath + 'ShellNew\' + TMakeNewFileActionItem(AItem).FileName;
						if not FileExists(SourceFileName) then
						begin
							SourceFileName := TMakeNewFileActionItem(AItem).FileName;
						end;
					end;
				end;
				if not CopyFileW(PWideChar(SourceFileName), PWideChar(NewFileName), True) then
					TMessageBox.Execute(SourceFileName + '𕡐ł܂', Text, MB_ICONERROR);
			end;
			else //nfkNullFile:
				FileClose(SysUtilsW.FileCreate(NewFileName));
			end;
		finally
			View.Model.ResumeSurveillance(True); {m҂}
		end;
		View.FocusTo(NewName);
	end
end;

procedure TMakeNewFileAction.DoUpdateItems(AItems: TActionItems);
var
	RootReg: TRegistryHive;

	function Proc(const NewFileExt: WideString): Boolean; cdecl;
	var
		Reg: TRegistryKey;
		FileType, FileDisp, ShellNewPath, NewFileName: WideString;
		Newable: Boolean;
		NewKind: TMakeNewFileKind;
	begin
		if NewFileExt[1] = '.' then
			try
				{^ʖڂɏĂlXg̗łꂱ ^ff̂悤ɂȂ̂
				OŊȗ, Xs[h͂Ȃ藎}

				Reg := TRegistryKey.Create(RootReg, NewFileExt, RegKeyRead);
				try
					FileType := Reg.ReadString('')
				finally
					Reg.Free
				end;
				if (FileType = '') or not RootReg.KeyExists(FileType) then Abort; {SEH}

				ShellNewPath := NewFileExt + '\' + FileType + '\ShellNew';
				Newable := RootReg.KeyExists(ShellNewPath);
				if not Newable then
				begin
					ShellNewPath := NewFileExt + '\ShellNew';
					Newable := RootReg.KeyExists(ShellNewPath);
				end;
				if not Newable then Abort; {SEH}

				Reg := TRegistryKey.Create(RootReg, ShellNewPath, RegKeyRead);
				try
					if Reg.ValueExists('NullFile') then
					begin
						NewKind := nfkNullFile;
						NewFileName := '';
					end
					else if Reg.ValueExists('FileName') then
					begin
						NewKind := nfkFileName;
						NewFileName := Reg.ReadString('FileName')
					end
					else
					begin
						Abort; {SEH}
						NewKind := nfkFileName; {xh~. SEH gĂƊm...}
					end;
				finally
					Reg.Free
				end;

				Reg := TRegistryKey.Create(RootReg, FileType, RegKeyRead);
				try
					FileDisp := Reg.ReadString('')
				finally
					Reg.Free
				end;
				if FileDisp = '' then FileDisp := FileType;

				with TMakeNewFileActionItem.Create(AItems, nil) do
				begin
					Text := FileDisp + '(' + NewFileExt + ')';
					FileExt := NewFileExt;
					Kind := NewKind;
					FileName := NewFileName;
				end
			except
			on EAbort do begin end; {Abort󂯎~߂}
			end;

		Result := False;
	end;

begin
	if FItems.Count = 0 then
	begin
		Application.BeginWrite;
		try
			if FItems.Count = 0 then {_u`FbNƂz}
			begin
				FItems.Title := 'VK쐬';
				with TMakeNewFileActionItem.Create(AItems, nil) do
				begin
					Text := '';
					FileExt := '';
					Kind := nfkNullFile;
				end;
				RootReg := TRegistryHive.Create(HKEY_CLASSES_ROOT);
				RootReg.Owner := Garbage;
				RootReg.EnumerateKeys(CEachName(@Proc), GetContext);
			end
		finally
			Application.EndWrite;
		end
	end
end;

function TMakeNewFileAction.GetItems: IActionItems;
begin
	Result := FItems
end;

procedure TMakeNewFileAction.Update;
var
	View: TFilerView;
begin
	View := ActiveView;
	SetEnabled((View <> nil) and (fpPlaceFile in View.Model.Possibilities));
end;

{ TWindowMenuAction }

constructor TWindowMenuAction.Create;
begin
	inherited Create(True);

	FItems := TActionItems.Create(Self);
	FItems.Owner := Self;

	SetName('WindowMenu');
	SetText('Window');
end;

procedure TWindowMenuAction.DoExecute;
begin
	TMessageBox.Execute('̃ANV̓j[pł', Text, MB_ICONWARNING)
end;

procedure TWindowMenuAction.DoExecuteItem(AItem: TActionItem);
var
	View: HWND;
begin
	View := HWND(AItem.Data);
	PostMessageW(GetParent(View), WM_APP_SETFOCUS, View, 0);
end;

procedure TWindowMenuAction.DoUpdateItems(AItems: TActionItems);

	function EnumWindow(W: TFilerWindow): Boolean cdecl;

		function EnumView(V: TFilerView): Boolean; cdecl;
		var
			Item: TActionItem;
			P: Integer;
		begin
			Item := TActionItem.Create(FItems, nil);
			Item.Data := Pointer(V.Handle);

			Item.Text := TFilerView(V).Model.DisplayText;
			P := AnsiPos('::{', Item.Text);
			if P > 0 then Item.Text := Copy(Item.Text, 1, P - 1);

			if (GetForegroundWindow = W.Handle) and (W.ActiveControl = V) then
				Item.Checked := True;

			Result := False
		end;

	begin
		W.EnumerateViews(CEachFilerView(@EnumView), GetContext);
		Result := False;
	end;

begin
	FItems.Clear;
	EnumerateFilerWindows(CEachFilerWindow(@EnumWindow), GetContext);
end;

function TWindowMenuAction.GetItems: IActionItems;
begin
	Result := FItems
end;

procedure TWindowMenuAction.Update;
begin
	FItems.PostUpdate
end;

{ TExtensionMenuAction }

type
	TDialogThread = class(T0Thread)
	private
    FExtensionInfo: IExtensionInfo;
	protected
		procedure Execute; override;
	public
		constructor Create(const AData: IExtensionInfo);
	end;

constructor TDialogThread.Create(const AData: IExtensionInfo);
begin
	FExtensionInfo := AData;
	inherited Create;
end;

procedure TDialogThread.Execute;
begin
	FExtensionInfo.Dialog
end;

constructor TExtensionMenuAction.Create;
begin
	inherited Create(True);

	FItems := TActionItems.Create(Self);
	FItems.Owner := Self;

	SetName('HelpMenu');
	SetText('Help');
end;

procedure TExtensionMenuAction.DoExecute;
begin
	TMessageBox.Execute('̃ANV̓j[pł', Text, MB_ICONWARNING)
end;

procedure TExtensionMenuAction.DoExecuteItem(AItem: TActionItem);
begin
	TDialogThread.Create(IExtensionInfo(AItem.Data));
end;

procedure TExtensionMenuAction.DoUpdateItems(AItems: TActionItems);
begin
end;

function TExtensionMenuAction.GetItems: IActionItems;
begin
	Result := FItems
end;

procedure TExtensionMenuAction.Update;

	function Enum(const E: IExtension): Boolean; cdecl;
	var
		Item: TActionItem;
		Info: IExtensionInfo;
	begin
		if Supports(E, IExtensionInfo, Info) then
		begin
			Item := TActionItem.Create(FItems, nil);
			Item.Text := Info.Name;
			Item.Data := Pointer(Info);
		end;
		Result := False
	end;

begin
	if FItems.Count = 0 then
		(Application as IExtensionManager).Enumerate(CEachExtension(@Enum), GetContext);
end;

end.

