unit UFileOp;

{ EOFile by YT }

interface

uses
	Windows, Messages, Types, SysUtils, ShellAPI, ActiveX,
	SysUtilsW, WinTips, ShlTips,
	LObjects, LActions, LGraphics, LWindows,
	LStdCtls, LCmmDlgs, LAddDlgs, LCmmCtls,
	UPath, UFolders,
	EOFileIntf;

type
	TFileOperationKind = (fokCopy, fokMove, fokLink);

const
	FileOperationKindText: array[TFileOperationKind] of WideString = ('Copy', 'Move', 'Link');

type
	TFileDuplicated = (fdOverwrite, fdBackup, fdRecycle,
					fdTimeStampOverwrite, fdTimeStampBackup, fdTimeStampRecyecle,
					fdRename, fdNone);

const
	FileDuplicatedText: array[TFileDuplicated] of WideString = (
					'㏑', 'obNAbv', 't@CS~',
					'㏑(^CX^vr)', 'obNAbv(^CX^vr)',
					't@CS~(^CX^vr)',
					'Oς', 'Ȃ');

type
	TFileOperationThread = class(T0GUIThread)
	private
		FKind: TFileOperationKind;
		FFromPath: TPath;
		FToPath: TPath;
		FFiles: TWideFileNameArray;
		FTotal: Integer;
		FFromArchive: Boolean;
	protected
		procedure Finalize; override;
		procedure Initialize; override;
		procedure Scan;
		procedure Process;
	public
		constructor Create(Kind: TFileOperationKind;
						const FromPath, ToPath: TPath; const Files: TWideFileNameArray);
	end;

implementation

function MakeShortCut(const FromFile, ToFile: TWideFileName): Boolean;
var
	ShellLink: TShellLink;
begin
	try
		ShellLink.Initialize;
		try
			ShellLink.Path := FromFile;
      ShellLink.WorkingDirectory := ExtractFilePath(FromFile);
			ShellLink.SaveToFile(ToFile);
		finally
			ShellLink.Finalize;
		end;
		Result := True;
	except
		Result := False;
	end
end;

function BackupName(const F: TWideFileName; Index: Integer): TWideFileName;
begin
	Result := WideFormat('%s.%.3d%s', [ChangeFileExt(F, ''), Index, ExtractFileExt(F)])
end;

procedure TimerProc(IDEvent, uReserved, dwUser, dwReserved1, dwReserved2: UINT); stdcall;
begin
  PostQuitMessage(0);
end;

type
	TFileOperationWindow = class(T0MainWindow)
	private
		FGauge: TProgressBar;
		FOptionGroup: TGroupBox;
		FOptions: array[TFileDuplicated] of TRadioButton;
		FAll: TCheckBox;
		FOk: TButton;
		FCancel: TButton;
		FProcessingFileName: TWideFileName;
		FQueryResult: Boolean;
		FSelected: TFileDuplicated;
		FForAll: Boolean;
		FClosed: Boolean;
		procedure SetProcessingFileName(const Value: TWideFileName);
		procedure WMSysCommand(var Message: TWMSysCommand); message WM_SYSCOMMAND;
	protected
		procedure Initialize; override;
		procedure PaintWindow(DC: HDC); override;
		constructor Create;
		procedure Setup; override;
		procedure SetMode(Dialoging: Boolean);
		procedure HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT); override;
		procedure DialogKey(var Message: TWMKey); override;
	public
		property ProcessingFileName: TWideFileName
						read FProcessingFileName write SetProcessingFileName;
		procedure Advance(Quick: Boolean);
		function Query(const FileName: WideString): Boolean;
		property Selected: TFileDuplicated read FSelected;
		property ForAll: Boolean read FForAll;
		function Closed: Boolean;
	end;

{ TFileOperationWindow }

procedure TFileOperationWindow.Advance(Quick: Boolean);
begin
	FGauge.Value := FGauge.Value + 1;
	Repaint;
	if not Quick then UpdateWindow(Handle);
end;

function TFileOperationWindow.Closed: Boolean;
var
	MSG: TMSG;
	TimerID: DWORD;
begin
	{鏈}
	TimerID := SetTimer(0, $3456, 1, @TimerProc);
	MessageLoop;
	KillTimer(0, TimerID);
	while PeekMessageW(MSG, 0, WM_TIMER, WM_TIMER, PM_REMOVE) do ;
	while PeekMessageW(MSG, 0, WM_QUIT, WM_QUIT, PM_REMOVE) do ;
	Result := FClosed
end;

constructor TFileOperationWindow.Create;
var
	CS: PCreateStructW;
begin
	inherited Create('', DefaultPoint, ZeroSize);
	CS := CreateStruct;
	CS^.style := CS^.style and not (WS_MINIMIZEBOX or WS_MAXIMIZEBOX or WS_THICKFRAME);
	CS^.dwExStyle := CS^.dwExStyle or (WS_EX_TOOLWINDOW or WS_EX_TOPMOST);
end;

procedure TFileOperationWindow.DialogKey(var Message: TWMKey);
begin
	if Message.CharCode = VK_ESCAPE then
	begin
		FQueryResult := False;
		PostQuitMessage(0);
	end
	else
		inherited;
end;

procedure TFileOperationWindow.HandleNotification(Sender: TAnyWindow; Info: PNMHDR; var Result: LRESULT);
begin
	if (Info.hwndFrom = FOk.Handle) and (Info.code = NM_COMMAND + BN_CLICKED) then
	begin
		FQueryResult := True;
		PostQuitMessage(0);
	end
	else if (Info.hwndFrom = FCancel.Handle) and (Info.code = NM_COMMAND + BN_CLICKED) then
	begin
		FQueryResult := False;
		PostQuitMessage(0);
	end
	else
		inherited;
end;

procedure TFileOperationWindow.Initialize;
var
	I: TFileDuplicated;
	SysMenu: HMENU;
begin
	inherited;

	FGauge := TProgressBar.Create(Self, Point(8, 72), Size(305, 9));

	FOptionGroup := TGroupBox.Create(Self, '', Point(4, 90), Size(313, 188));

	FAll := TCheckBox.Create(FOptionGroup, 'ȌS(&A)', Point(200, 20), Size(100, 20));

	for I := Low(I) to High(I) do
		FOptions[I] := TRadioButton.Create(FOptionGroup, FileDuplicatedText[I], Point(8, Ord(I) * 20 + 20), Size(290, 20));

	FOptions[fdBackup].Checked := True;
	FSelected := fdBackup;

	FOk := TButton.Create(Self, 'OK', Point(320 - 150 - 12, FOptionGroup.BoundsRect.Bottom + 7), DefaultSize);
	FOk.IsDefault := True;

	FCancel := TButton.Create(Self, 'Cancel', Point(320 - 75 - 8, FOptionGroup.BoundsRect.Bottom + 7), DefaultSize);

	SysMenu := GetSystemMenu(Handle, False);
	DeleteMenu(SysMenu, SC_TASKLIST, MF_BYCOMMAND);
	DeleteMenu(SysMenu, 7, MF_BYPOSITION);
	DeleteMenu(SysMenu, 5, MF_BYPOSITION);
	DeleteMenu(SysMenu, SC_MAXIMIZE, MF_BYCOMMAND);
	DeleteMenu(SysMenu, SC_MINIMIZE, MF_BYCOMMAND);
	DeleteMenu(SysMenu, SC_SIZE, MF_BYCOMMAND);
	DeleteMenu(SysMenu, SC_RESTORE, MF_BYCOMMAND);

	SetMode(False);
end;

procedure TFileOperationWindow.PaintWindow(DC: HDC);
begin
	with TFileOperationThread(Owner) do
	begin
		SelectObject(DC, DefaultGUIFont.Handle);
		SetBkMode(DC, TRANSPARENT);
		TextOutW(DC, 8,  8, 'From:', 5);
		TextOutW(DC, 8, 28, 'To:', 3);
		TextOut(DC, 48,  8, MinimizeName(FFromPath.Directory, DC, 265));
		TextOut(DC, 48, 28, MinimizeName(FToPath.Directory, DC, 265));

		TextOut(DC, 8, 48, FProcessingFileName);

		SetTextAlign(DC, TA_RIGHT);
		if FTotal = 0 then
			TextOut(DC, 312, 48, 'Scanning...')
		else if FTotal = FGauge.Value then
			TextOut(DC, 312, 48, 'Completed')
		else
			TextOut(DC, 312, 48, WideFormat('%d / %d', [FGauge.Value, FTotal]));
	end
end;

function TFileOperationWindow.Query(const FileName: WideString): Boolean;
var
	I: TFileDuplicated;
begin
	if not FForAll then
	begin
		FQueryResult := False;

		FOptionGroup.Text := FileName + '͊ɑ݂Ă܂';
		SetMode(True);
		MessageLoop;
		SetMode(False);
		UpdateWindow(Handle);

		FForAll := FAll.Checked;

		FSelected := fdNone;
		for I := Low(I) to High(I) do
			if FOptions[I].Checked then
			begin
				FSelected := I;
				Break;
			end;
	end;
	Result := FQueryResult;
end;

procedure TFileOperationWindow.SetMode(Dialoging: Boolean);
var
	S: TSize;
begin
	S.cx := 320;
	S.cy := 88;
	if Dialoging then S.cy := FCancel.BoundsRect.Bottom + 3;

	ClientSize := S;
	FOptionGroup.Enabled := Dialoging;
	FOk.Enabled := Dialoging;
	FCancel.Enabled := Dialoging;

	if Dialoging then
		FOptions[FSelected].SetFocus;
end;

procedure TFileOperationWindow.SetProcessingFileName(
	const Value: TWideFileName);
begin
	FProcessingFileName := Value;
	Repaint;
	UpdateWindow(Handle)
end;

procedure TFileOperationWindow.Setup;
begin
	inherited;
	Text := FileOperationKindText[TFileOperationThread(Owner).FKind];

	UpdateWindow(Handle);
	TFileOperationThread(Owner).Scan;
	FGauge.Max := TFileOperationThread(Owner).FTotal;
	FQueryResult := True;
	TFileOperationThread(Owner).Process;
	if FQueryResult then
	begin
    Beep;
		Application.Async;
		Sleep(1000);
		Application.RestoreAsync;
	end;
	{^MessageLoop𖳌}
	PostQuitMessage(0);
end;

procedure TFileOperationWindow.WMSysCommand(var Message: TWMSysCommand);
begin
	if Message.CmdType = SC_CLOSE then FClosed := True;
	inherited;
end;

{ TFileOperationThread }

constructor TFileOperationThread.Create(Kind: TFileOperationKind;
	const FromPath, ToPath: TPath; const Files: TWideFileNameArray);
var
	F: TWin32FindDataW;
	H: THandle;
begin
	inherited Create;
	FKind := Kind;
	FFromPath := CopyPath(FromPath);
	FToPath := CopyPath(ToPath);
	FFiles := Files;

	FFromArchive := FromPath.Archive <> '';
	if not FFromArchive then
	begin
		H := FindFirstFileW(PWideChar(FromPath.Directory), F);
		Windows.FindClose(H);
		FFromArchive := (H <> INVALID_HANDLE_VALUE) and (F.dwFileAttributes and faDirectory = 0);
	end;
end;

procedure TFileOperationThread.Finalize;
begin
	FinalizePath(FFromPath);
	FinalizePath(FToPath);
	inherited;
	CoUninitialize;
end;

procedure TFileOperationThread.Initialize;
begin
	inherited;
	CoInitialize(nil);
	TFileOperationWindow.Create.Owner := Self
end;

procedure TFileOperationThread.Process;
var
	InputDialog: TInputDialog;
	SamePlace: Boolean;
	Manager: IFolderManager;

	procedure Exec(const FromPath, ToPath: TPath; const AFile: PFile; const SourceArchive: IArchive; CountOnly: Boolean);
		{SourceArchive ... FFromArchive = True ̎̂ݎgp}
	var
		ToFile: TWideFileName;
		FromFile: TWideFileName; {DirtpXAArc\}

		procedure Backup;
		const
			CopyOrRename: array[Boolean] of DWORD = (FO_RENAME, FO_COPY);
		var
			I, NextNumber: Integer;
			Bk: WideString;
		begin
			NextNumber := 0;
			for I := 0 to 998 do
			begin
				Bk := BackupName(ToFile, I);
				if FileSystemExists(Bk) then
				begin
					if I <> NextNumber then
					begin
						ShellCopyOrMoveOrRenameFile(MainWindow.Handle,
										Bk, BackupName(ToFile, NextNumber), FO_RENAME);
					end;
					Inc(NextNumber);
				end;
			end;
			Bk := BackupName(ToFile, NextNumber);
			ShellCopyOrMoveOrRenameFile(MainWindow.Handle, ToFile, Bk, CopyOrRename[SamePlace]);
		end;

		procedure Recycle;
		begin
			if SamePlace then
				TMessageBox.Execute(MainWindow, 'ꏊōƒɃS~͂ł܂', FileOperationKindText[FKind], MB_ICONWARNING)
			else
				ShellDeleteFiles(MainWindow.Handle, '', [ToFile], FOF_ALLOWUNDO or FOF_SILENT or FOF_NOCONFIRMATION);
		end;

		function CompareTimeStamp: Boolean;
		var
			TD: TWin32FindDataW;
		begin
			Windows.FindClose(FindFirstFileW(PWideChar(ToFile), TD));
			Result := CompareFileTime(AFile^.LastWriteTime, TD.ftLastWriteTime) > 0
		end;

	var
		Succeed, ForOperating, CopyTimeStamp, SubCountOnly: Boolean;
		H: THandle;
		FD: WIN32_FIND_DATAW;
		I: Integer;
		FromPath2, ToPath2: TPath;
		File2: TFile;
		Arc2: IArchive;
		AFile_Name: WideString;
	begin
		{̃t@CƂ\}
		if not CountOnly then
			TFileOperationWindow(MainWindow).ProcessingFileName := AFile.DisplayName;

		{tpXgݗĂ}
		if FFromArchive then
		begin
			FromFile := FromPath.Directory + '/';
			if FromPath.Archive <> '' then FromFile := FromFile + FromPath.Archive + '/';
			FromFile := FromFile + AFile.Name;
		end
		else
		begin
			FromFile := ConcatPath(FromPath.Directory, AFile.Name);
		end;
		SetString(AFile_Name, AFile^.Name, AFile^.NameLength);
		ToFile := ConcatPath(ToPath.Directory, AFile_Name);

		{܂œKv邩H}
		if (FKind <> fokLink) and (AFile.Attributes and faDirectory <> 0) then
		begin
			if CompareFileName(FromFile + '\', Copy(ToFile + '\', 1, Length(FromFile) + 1)) = 0 then
			begin
				TMessageBox.Execute(MainWindow, FromFile + #13#10 + ToFile + #13#10 +
								'Rs[̎qRs[ɂƖRs[ɂȂ܂̂Œf.',
								FileOperationKindText[FKind], MB_ICONWARNING);
				Exit;
			end;

			SubCountOnly := CountOnly or
				(FKind = fokMove) and not DirectoryExists(ToFile) and
					(ExtractFileDrive(ToFile) = ExtractFileDrive(FromFile));

			CopyTimeStamp := not SubCountOnly and
				not SamePlace and CreateDirectoryW(PWideChar(ToFile), nil);

			if FFromArchive then
			begin
				FromPath2 := FromPath;
				if FromPath2.Archive <> '' then FromPath2.Archive := FromPath2.Archive + '/';
				FromPath2.Archive := FromPath2.Archive + AFile_Name;
				ToPath2 := ToPath;
				ToPath2.Directory := ToFile;
				Application.BeginWrite;
				try
					Arc2 := Manager.GetFolder(FromPath2) as IArchive;
				finally
					Application.EndWrite;
				end;
				for I := 0 to Arc2.Count - 1 do
					Exec(FromPath2, ToPath2, Arc2.Items[I], Arc2, CountOnly);
			end
			else
			begin
				{ċAIɃRs[... Dir ̉Ȃ̂Ŗⓚp \ t}
				H := FindFirstFileW(PWideChar(FromFile + '\*.*'), FD);
				if H = INVALID_HANDLE_VALUE then RaiseLastOSError;
				repeat
					if (FD.cFileName[0] <> '.') or (FD.cFileName[1] <> '.') and (FD.cFileName[1] <> #0) then
					begin
						FromPath2 := FromPath;
						FromPath2.Directory := FromFile;
						ToPath2 := ToPath;
						ToPath2.Directory := ToFile;
						FindDataToFile(File2, FD, Manager);
						Exec(FromPath2, ToPath2, @File2, nil, SubCountOnly);
					end
				until not FindNextFileW(H, FD);
				Windows.FindClose(H);

				{Ƀt@CcĂꍇRemoveDirectory API͍pȂ}
				if (FKind = fokMove) and not SubCountOnly then
					RemoveDirectoryW(PWideChar(FromFile));
			end;

			if not CountOnly and SubCountOnly then
			begin
				Application.Async;
				try
					Succeed := ShellCopyOrMoveOrRenameFile(MainWindow.Handle, FromFile, ToFile, FO_MOVE);
					if not Succeed then
						if TMessageBox.Execute(MainWindow, FromFile + '̏Ɏs. ܂?',
										FileOperationKindText[FKind], MB_YESNO or MB_DEFBUTTON2 or MB_ICONERROR) = IDNO then
							Abort;
				finally
					Application.RestoreAsync;
				end;
			end
			else if CopyTimeStamp then
			begin
				{^CX^ṽRs[NTn̂ݗLc2000verł͑OȂ̂OK}
(*			if not FFromArchive then
				begin
					H := CreateFileW(PWideChar(FromFile), GENERIC_READ,
								FILE_SHARE_READ or FILE_SHARE_WRITE,
								nil, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0);
					if H = INVALID_HANDLE_VALUE then RaiseLastOSError;
					GetFileTime(H, @CT, @AT, @WT);
					CloseHandle(H);
				end *)
				H := CreateFileW(PWideChar(ToFile), GENERIC_READ or GENERIC_WRITE,
								FILE_SHARE_READ or FILE_SHARE_WRITE,
								nil, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, 0);
				if H = INVALID_HANDLE_VALUE then RaiseLastOSError;
				with AFile^ do SetFileTime(H, @CreationTime, @LastAccessTime, @LastWriteTime);
				CloseHandle(H);
			end;
		end
		else if not CountOnly then
		begin
			{V[gJbg쐬̊gq׍H}
			if FKind = fokLink then
			begin
				if CompareFileName(ExtractFileExt(ToFile), '.exe') = 0 then
					ToFile := ChangeFileExt(ToFile, '');

				ToFile := ToFile + '.lnk';
			end;

			{Rs[肪܂Ń[v}
			ForOperating := True;
			while FileSystemExists(ToFile) do
			begin
				if not TFileOperationWindow(MainWindow).Query(ToFile) then Abort;

				case TFileOperationWindow(MainWindow).Selected of
				fdOverwrite:
				begin
					ForOperating := not SamePlace;
					Break;
				end;
				fdBackup:
				begin
					Backup;
					ForOperating := not SamePlace;
					Break;
				end;
				fdRecycle:
				begin
					Recycle;
					ForOperating := True;
				end;
				fdTimeStampOverwrite:
				begin
					ForOperating := CompareTimeStamp;
					Break;
				end;
				fdTimeStampBackup:
				begin
					ForOperating := CompareTimeStamp;
					if ForOperating then Backup;
					Break
				end;
				fdTimeStampRecyecle:
				begin
					ForOperating := CompareTimeStamp;
					if ForOperating then Recycle;
					Break
				end;
				fdRename:
				begin
					ForOperating := True;
					InputDialog.Text := AFile.Name;
					InputDialog.Value := ExtractFileName(ToFile);
					InputDialog.SelectAll;
					if InputDialog.Execute then
						ToFile := ExtractFilePath(ToFile) + InputDialog.Value;
				end;
				else
					ForOperating := False;
					Break;
				end;
			end;
			if ForOperating then
			begin
				Application.Async;
				try
					case FKind of
					fokCopy:
					begin
						if FFromArchive then
							try
								SourceArchive.Unpack(AFile, ToPath.Directory);
								Succeed := True;
							except
								ShowException(ExceptObject, ExceptAddr);
								Succeed := False;
							end
						else
						begin
							Succeed := ShellCopyOrMoveOrRenameFile(MainWindow.Handle, FromFile, ToFile, FO_COPY);
						end
					end;
					fokMove:
					begin
						Succeed := ShellCopyOrMoveOrRenameFile(MainWindow.Handle, FromFile, ToFile, FO_MOVE);
					end
					else
						Succeed := MakeShortCut(FromFile, ToFile);
					end;
					if not Succeed then
						if TMessageBox.Execute(MainWindow, FromFile + '̏Ɏs. ܂?',
										FileOperationKindText[FKind], MB_YESNO or MB_DEFBUTTON2 or MB_ICONERROR) = IDNO then
							Abort;
				finally
					Application.RestoreAsync;
				end;
			end;
		end;
		TFileOperationWindow(MainWindow).Advance(CountOnly);
		if not CountOnly and TFileOperationWindow(MainWindow).Closed then Abort;
	end;

var
	FromPath2: TPath;
	FileName: TWideFileName;
	I, J: Integer;
	Archive: IArchive;
	File2: PFile;
	File2Body: TFile;
	FD: TWin32FindDataW;
begin
	Manager := Application as IFolderManager;

	SamePlace := ComparePath(FFromPath, FToPath);

	{OύXp̓̓_CAO邾Ă}
	InputDialog := TInputDialog.Create(MainWindow, '', False);
	InputDialog.Owner := Garbage;

	try
		for I := Low(FFiles) to High(FFiles) do
		begin
			FromPath2 := FFromPath;
			FileName := FFiles[I];
			if Pos('\', FileName) > 0 then
			begin
				InitializePath(FromPath2);
				FromPath2.Directory := ExtractFileDir(FileName);
				FileName := ExtractFileName(FileName);
			end;

			if FFromArchive then
			begin
				Archive := Manager.GetFolder(FromPath2) as IArchive;
				File2 := nil;
				for J := Archive.Count - 1 downto 0 do
				begin
					File2 := Archive.Items[J];
					if CompareFileName(File2.Name, File2.NameLength, PWideChar(FileName), Length(FileName)) = 0 then
						Break;
				end;
			end
			else
			begin
				Windows.FindClose(FindFirstFileW(PWideChar(ConcatPath(FromPath2.Directory, FileName)), FD));
				FindDataToFile(File2Body, FD, Manager);
				File2 := @File2Body;
			end;
			Exec(FromPath2, FToPath, File2, Archive, False)
		end
	except
	on EAbort do
		begin end;
	else
		ShowException(ExceptObject, ExceptAddr)
	end;
	{m̂Update͕sv}
end;

procedure TFileOperationThread.Scan;

	procedure ScanDir(const Path: TPath);
	var
		FD: WIN32_FIND_DATAW;
		H: THandle;
		Path2: TPath;
	begin
		H := FindFirstFileW(PWideChar(Path.Directory + '\*'), FD);
		if H <> INVALID_HANDLE_VALUE then
		begin
			repeat
				if (FD.cFileName[0] <> '.') or (FD.cFileName[1] <> '.') and (FD.cFileName[1] <> #0) then
				begin
					Inc(FTotal);
					Path2 := Path;
					Path2.Directory := Path2.Directory + '\' + FD.cFileName;
					ScanDir(Path2);
				end;
			until not FindNextFileW(H, FD);
			Windows.FindClose(H);
		end
	end;

	procedure ScanArc(const Path: TPath);
	var
		Path2: TPath;
		Folder: IFolder;
		I: Integer;
		Item: PFile;
		FileName: WideString;
	begin
		Folder := (Application as IFolderManager).GetFolder(Path);
		if Folder.Valid then
		begin
			for I := Folder.Count - 1 downto 0 do
			begin
				Inc(FTotal);
				Item := Folder.Items[I];
				if foEnter in Item.Operatable then
				begin
					with Item^ do SetString(FileName, Name, NameLength);
					Path2 := Path;
					Path2.Archive := Path2.Archive + '/' + FileName;
					ScanArc(Path2);
				end;
			end;
		end
	end;

var
	I: Integer;
	Path2: TPath;
begin
	FTotal := Length(FFiles);
	if FKind <> fokLink then
		if not FFromArchive then
		begin
			for I := High(FFiles) downto Low(FFiles) do
			begin
				Path2 := FFromPath;
				Path2.Directory := Path2.Directory + '\' + FFiles[I];
				ScanDir(Path2);
			end;
		end
		else
		begin
			for I := High(FFiles) downto Low(FFiles) do
			begin
				Path2 := FFromPath;
				if Path2.Archive <> '' then Path2.Archive := Path2.Archive + '/';
				Path2.Archive := Path2.Archive + FFiles[I];
				ScanArc(Path2);
			end;
		end
end;

end.
