unit UFolders;

{ EOFile by YT }

interface

uses
	Windows, Types, SysUtils, ActiveX, ShlObj, SysConst,
	SysUtilsW, SysHacks, AlgUtils, WinTips, ShlTips,
	LConsts, LObjects, LWindows,
	EOFileIntf, UPath;

type
	TWideFileNameArray = TWideStringDynArray;

	TFileArray = array of TFile;
	TFilePtrArray = array of PFile;

type
	TFolderManager = class(T0Object, IFolderManager, IObserver)
	private
		FFolders: array of IFolder;
		procedure Observe;
		{IFolderManager}
		function GetFolder(const APath: TPath): IFolder;
		function GetOpenedFolder(const APath: TPath): IFolder;
		procedure Refresh(const APath: TPath; Sub: Boolean);
		procedure ChangePath(const APath: TPath; const NewName: WideString);
		function FindArchiver(const Ext: WideString): IArchiver;
		function LinkToPath(const LinkFile: WideString): TPath;
		function ToPath(const Path: WideString): TPath;
		function ToString(const Path: TPath): WideString;
	public
		procedure Pack;
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
	end;

type
	T0Folder = class(TInterfacedObject, IObservable)
	private
		FPath: TPath;
		FRefreshed: THandle;
		FWatchThread: T0Thread;
		FViews: TObservable;
	protected
		{IObservable}
		procedure Attach(const AView: IObserver);
		procedure Detach(const AView: IObserver);
		function HasClients: Boolean;
		{IFolder}
		function GetValid: Boolean;
		function GetItems(AIndex: Integer): PFile;
		function GetCount: Integer;
		function GetPath: TPath;
		procedure StopSurveillance;
		procedure ResumeSurveillance(Wait: Boolean);
		function GetInfoText(AFile: PFile): WideString;
	protected
		FFiles: TFileArray;
		FValid: Boolean;
		procedure DoRefresh; virtual; abstract;
		property Views: TObservable read FViews;
		procedure ViewDecreasedAll; virtual;
	public
		constructor Create(const APath: TPath);
		procedure BeforeDestruction; override;
		procedure Refresh;
		property Valid: Boolean read FValid;
		property Items[AIndex: Integer]: PFile read GetItems;
		property Count: Integer read GetCount;
		property Path: TPath read FPath;
	end;

	TDirectory = class(T0Folder, IFolder, IDirectory)
	private
		FAtRoot: Boolean;
		function GetDisplayText: WideString;
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		function GetPossibilities: TFolderPossibilities;
		procedure Move(const NewPath: WideString);
		function ConcatenatePath(AFile: PFile): TPath;
		{IDirectory}
		function GetAtRoot: Boolean;
	protected
		procedure DoRefresh; override;
		procedure BeginWatch;
		procedure EndWatch;
	public
		constructor Create(const ADirectory: TWideFileName);
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
	end;

	TNameSpace = class(T0Folder, IFolder, INameSpace, IShellFolder)
	private
		FShellFolder: IShellFolder;
		FDisplayName: WideString;
		FIsRecycleBox: Boolean;
		FIsDrives: Boolean;
		function GetDisplayText: WideString;
		function GetParentPath: TPath;
		function IsSub(const AParent: TPath): Boolean;
		function GetPossibilities: TFolderPossibilities;
		procedure Move(const NewPath: WideString);
		function ConcatenatePath(AFile: PFile): TPath;
		{INameSpace}
		function AtDesktop: Boolean;
		function AtDrives: Boolean;
	protected
		procedure DoRefresh; override;
	public
		constructor Create(AItemIDList: PItemIDList);
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;
		property ShellFolder: IShellFolder read FShellFolder implements IShellFolder;
	end;

procedure FindDataToFile(var Dest: TFile; const Data: TWin32FindDataW; const Manager: IFolderManager);
function FilePtrArray2WideFileNameArray(const A: TFilePtrArray): TWideFileNameArray;
procedure SetCurrentDirNull;
function DirectoryIsSub(const Dir: WideString; const Path: TPath): Boolean;
function GetParentFolder(const Folder: IFolder): IFolder;

var
	DrivesPath, BitbucketPath: TWideFileName;

function IsDrives(const Model: IFolder): Boolean;
function IsRootDir(const Model: IFolder): Boolean;

implementation

function IsDrives(const Model: IFolder): Boolean;
var
	NameSpace: INameSpace;
begin
	Result := Supports(Model, INameSpace, NameSpace) and NameSpace.AtDrives
end;

function IsRootDir(const Model: IFolder): Boolean;
var
	Dir: IDirectory;
begin
	Result := Supports(Model, IDirectory, Dir) and Dir.AtRoot
end;

function FilePtrArray2WideFileNameArray(const A: TFilePtrArray): TWideFileNameArray;
var
	I: Integer;
begin
	SetLength(Result, Length(A));
	for I := High(A) downto Low(A) do
		SetString(Result[I], A[I].Name, A[I].NameLength);
end;

procedure SetCurrentDirNull;
var
	P: array[0..MAX_PATH] of WideChar;
begin
	GetTempPathW(MAX_PATH, P);
	if not SetCurrentDirectoryW(P) then RaiseLastOSError;
end;

function SpecialFolderPath(Folder: Integer): TWideFileName;
var
	IDList: PItemIDList;
	Desktop: IShellFolder;
begin
	if FAILED(SHGetSpecialFolderLocation(0, Folder, IDList)) then
		RaiseLastOSError;

	if IDList.mkid.cb = 0 then
		Result := ''
	else
	begin
		SHGetDesktopFolder(Desktop);
		Result := GetDisplayName(Desktop, IDList, SHGDN_FORPARSING);
	end;

	CoTaskMemFree(IDList);
end;

procedure FindDataToFile(var Dest: TFile; const Data: TWin32FindDataW; const Manager: IFolderManager);
begin
	Dest.Kind := fkFile;

	Dest.NameLength := lstrlenW(Data.cFileName);
	//lstrcpyW(Dest.Name, Data.cFileName);
	Move(Data.cFileName, Dest.Name, (Dest.NameLength + 1) * SizeOf(WideChar));

	Dest.DisplayName := Dest.Name;
	Dest.DisplayNameLength := Dest.NameLength;

	Int64Rec(Dest.Size).Lo := Data.nFileSizeLow;
	Int64Rec(Dest.Size).Hi := Data.nFileSizeHigh;

  Dest.Number := High(Word);

	Dest.CreationTime := Data.ftCreationTime;
	Dest.LastWriteTime := Data.ftLastWriteTime;
	Dest.LastAccessTime := Data.ftLastAccessTime;

	Dest.Attributes := Data.dwFileAttributes;

	Dest.ShortNameLength := lstrlenW(Data.cAlternateFileName);
	//lstrcpyW(Dest.ShortName, Data.cAlternateFileName);
	Move(Data.cAlternateFileName, Dest.ShortName, (Dest.ShortNameLength + 1) * SizeOf(WideChar));

	Dest.Operatable := AllFileOperatable;
	if (Dest.Attributes and faDirectory = 0) and (Manager.FindArchiver(Dest.Name) = nil) then
		Exclude(Dest.Operatable, foEnter);
end;

function DirectoryIsSub(const Dir: WideString; const Path: TPath): Boolean;
var
	ParentDir: TWideFileName;
begin
	if Path.ItemIDList = nil then
	begin
		ParentDir := IncludeTrailingPathDelimiter(Path.Directory);
		Result := CompareFileName(ParentDir, Copy(Dir + '\', 1, Length(ParentDir))) = 0
	end
	else
	begin
		{fXNgbv}CRs[^}
		Result := (Path.ItemIDList.mkid.cb = 0) or (Path.Directory = DrivesPath);
	end;
end;

function GetParentFolder(const Folder: IFolder): IFolder;
var
	Path: TPath;
begin
	Path := Folder.GetParentPath;
	Result := (Application as IFolderManager).GetFolder(Path);
	FinalizePath(Path);
end;

type
	TDirWatchThread = class(T0Thread)
	private const
		WAIT_OBJECT_LOW = WAIT_OBJECT_0 + 1;
		WAIT_OBJECT_HIGH = WAIT_OBJECT_0 + 2;
	private
		FEvents: packed record
			Terminated: THandle;
			Watcher: array[WAIT_OBJECT_LOW..WAIT_OBJECT_HIGH] of THandle;
		end;
		FSurveillance: THandle;
		FPath: PPath;
		FCount: Integer;
	protected
		procedure Execute; override;
		procedure Finalize; override;
	public
		constructor Create(AFolder: T0Folder);
		procedure Join;
		property Surveillance: THandle read FSurveillance;
	end;

constructor TDirWatchThread.Create(AFolder: T0Folder);
const
	Notify = FILE_NOTIFY_CHANGE_FILE_NAME or
				FILE_NOTIFY_CHANGE_DIR_NAME or
				FILE_NOTIFY_CHANGE_ATTRIBUTES or
				FILE_NOTIFY_CHANGE_SIZE or
				FILE_NOTIFY_CHANGE_LAST_WRITE;
begin
	FEvents.Terminated := CreateEventW(nil, True, False, nil);
	{ZbgAlOFF}

	FPath := @AFolder.FPath;
	if FPath.ItemIDList = nil then
	begin
		FCount := 1;
		FEvents.Watcher[1] := FindFirstChangeNotificationW(PWideChar(FPath.Directory), False, Notify);
	end
	else
	begin
		FCount := 2;
		FEvents.Watcher[1] := FindFirstChangeNotificationW(PWideChar(SpecialFolderPath(CSIDL_DESKTOPDIRECTORY)), False, Notify);
		FEvents.Watcher[2] := FindFirstChangeNotificationW(PWideChar(SpecialFolderPath(CSIDL_COMMON_DESKTOPDIRECTORY)), False, Notify);
	end;

	FSurveillance := CreateEventW(nil, True, True, nil);
	{ZbgAlON}

	inherited Create;
end;

procedure TDirWatchThread.Finalize;
begin
	CloseHandle(FEvents.Terminated);

	FindCloseChangeNotification(FEvents.Watcher[1]);
	if FCount > 1 then
		FindCloseChangeNotification(FEvents.Watcher[2]);

	CloseHandle(FSurveillance);

	inherited;
end;

procedure TDirWatchThread.Execute;
label
	Exit;
var
	R: Integer;
begin
	//OutputDebugStringW(PWideChar(WideCurrentDirectory));

	repeat
		R := WaitForMultipleObjects(FCount + 1, Pointer(@FEvents), False, INFINITE);
		case R of
		WAIT_OBJECT_0 + 1, WAIT_OBJECT_0 + 2:
			;
		else
			goto Exit;
		end;

		{ύXSēǂݐ؂}
		repeat
			if not FindNextChangeNotification(FEvents.Watcher[R]) then goto Exit;
			case WaitForSingleObject(FEvents.Watcher[R], 250) of
			WAIT_TIMEOUT:
				Break;
			WAIT_OBJECT_0:
				;
			else
				goto Exit;
			end;
		until False;

		{܂ōǂ~߂Ă}
		if WaitForSingleObject(FSurveillance, INFINITE) <> WAIT_OBJECT_0 then goto Exit;

		Application.BeginWrite;
		try
			if WaitForSingleObject(FEvents.Terminated, 0) = WAIT_TIMEOUT then
				(Application as IFolderManager).Refresh(FPath^, False);
		finally
			Application.EndWrite;
		end;

	until not FindNextChangeNotification(FEvents.Watcher[R]);

Exit:
	WaitForSingleObject(FEvents.Terminated, INFINITE)
end;

procedure TDirWatchThread.Join;
begin
	SetEvent(FEvents.Terminated)
end;

{ TFolderManager }

procedure TFolderManager.AfterConstruction;
begin
	inherited;

	DrivesPath := SpecialFolderPath(CSIDL_DRIVES);
	BitbucketPath := SpecialFolderPath(CSIDL_BITBUCKET);

	Application.AsObservable.Attach(Self);
end;

procedure TFolderManager.BeforeDestruction;
begin
	Application.AsObservable.Detach(Self);
	Pack;
	inherited;
end;

function TFolderManager.GetOpenedFolder(const APath: TPath): IFolder;
var
	I: Integer;
begin
	Assert(APath.ItemIDList <> Pointer(1)); {̓[hł邩}

	for I := Low(FFolders) to High(FFolders) do
	begin
		if ComparePath(FFolders[I].Path, APath) then
		begin
			Result := FFolders[I];
			Exit;
		end;
	end;

	Result := nil
end;

function TFolderManager.GetFolder(const APath: TPath): IFolder;

	function PathToFolder(const APath: TPath): IFolder;
	var
		A: IArchiver;
		R: IFolder;

		function EachExtension(const Extension: IExtension): Boolean; cdecl;
		var
			Handler: IPathHandler;
		begin
			Result := Supports(Extension, IPathHandler, Handler);
			if Result then
			begin
				R := Handler.TryOpen(APath);
				Result := R <> nil;
			end;
		end;

	var
		IsDir: Boolean;
		E: UINT;
	begin
		if APath.Extension <> nil then
		begin
			(Application as IExtensionManager).Enumerate(CEachExtension(@EachExtension), GetContext);
			if R <> nil then
			begin
				Result := R;
				Exit;
			end;
		end;

		Assert(APath.Extension = nil);

		if APath.ItemIDList <> nil then
			Result := TNameSpace.Create(APath.ItemIDList)
		else
		begin
			E := SetErrorMode(SEM_FAILCRITICALERRORS);
			IsDir := GetFileAttributesW(Pointer(APath.Directory)) and faDirectory <> 0;
			SetErrorMode(E);
			if not IsDir then
			begin
				A := FindArchiver(APath.Directory);
				if A <> nil then
				begin
					Result := (A as IArchiver).Open(APath);
					Exit
				end;
			end;

			{Invalid𗘗p邱Ƃ̂ŊmF͂Ȃ}
			Result := TDirectory.Create(APath.Directory)
		end;
	end;

begin
	Result := GetOpenedFolder(APath);
	if Result = nil then
	begin
		Application.BeginWrite;
		try
			Result := PathToFolder(APath);
			SetLength(FFolders, Length(FFolders) + 1);
			FFolders[High(FFolders)] := Result;
		finally
			Application.EndWrite;
		end
	end;
end;

procedure TFolderManager.ChangePath(const APath: TPath; const NewName: WideString);
var
	I: Integer;
begin
	for I := Low(FFolders) to High(FFolders) do
	begin
		if ComparePath(FFolders[I].Path, APath) or FFolders[I].IsSub(APath) then
		begin
			FFolders[I].Move(NewName + Copy(FFolders[I].Path.Directory,
							Length(APath.Directory) + 1, MaxInt));
			FFolders[I].Refresh;
		end;
	end;
end;

procedure TFolderManager.Refresh(const APath: TPath; Sub: Boolean);
var
	I: Integer;
begin
	Assert(APath.ItemIDList <> Pointer(1)); {̓[hł邩}

	for I := Low(FFolders) to High(FFolders) do
	begin
{		if FixedOnly and not Supports(FFolders[I], IDirectory) then
			if GetDriveTypeW(PWideChar(WideExtractFileDrive(FFolders[I].Path.Directory) + '\')) <> DRIVE_FIXED then
				Continue; }

		if ComparePath(FFolders[I].Path, APath) or (Sub and FFolders[I].IsSub(APath)) then
			FFolders[I].Refresh;
	end
end;

procedure TFolderManager.Pack;
var
	I, Used: Integer;
begin
	Used := 0;
	for I := Low(FFolders) to High(FFolders) do
		if FFolders[I].HasClients then
		begin
			FFolders[Used] := FFolders[I];
			Inc(Used);
		end;

	SetLength(FFolders, Used);
end;

procedure TFolderManager.Observe;
begin
	Pack
end;

function TFolderManager.FindArchiver(const Ext: WideString): IArchiver;
var
	E: WideString;
	Context: Pointer;

	function EachArchiveKind(const Ext: WideString): Boolean; cdecl;
	begin
		Result := CompareFileName(E, Ext) = 0
	end;

	function EachExtension(const Extension: IExtension): Boolean; cdecl;
	var
		A: IArchiver;
	begin
		Result := Supports(Extension, IArchiver, A) and
			(A.Enumerate(CEachArchiveKind(@EachArchiveKind), Context) <> '')
	end;

begin
	E := ExtractFileExt(Ext);
	Context := GetContext;
	Result := (Application as IExtensionManager).
		Enumerate(CEachExtension(@EachExtension), Context) as IArchiver
end;

function TFolderManager.LinkToPath(const LinkFile: WideString): TPath;
var
	ShellLink: TShellLink;
begin
	InitializePath(Result);
	ShellLink.Initialize;
	try
		ShellLink.LoadFromFile(LinkFile);
		ShellLink.Resolve(CurrentThread.LeadWindow.Handle);

		Result.Directory := ExtractLongPathName(ShellLink.Path);
		if Result.Directory = '' then
			Result.ItemIDList := ShellLink.ItemIDList;
	finally
		ShellLink.Finalize
	end;
end;

function TFolderManager.ToPath(const Path: WideString): TPath;
begin
	Result := ParsePath(Path)
end;

function TFolderManager.ToString(const Path: TPath): WideString;
var
	Desktop: IShellFolder;
begin
	if Path.ItemIDList <> nil then
	begin
		SHGetDesktopFolder(Desktop);
		Result := GetDisplayName(Desktop, Path.ItemIDList, SHGDN_FORPARSING);
	end
	else if Path.Archive = '' then
	begin
		Result := ExtractLongPathName(Path.Directory)
	end
	else
	begin
		Result := ExtractLongPathName(Path.Directory) + '/' + Path.Archive
	end
end;

{ T0Folder }

constructor T0Folder.Create(const APath: TPath);
begin
	inherited Create;
	FPath := CopyPath(APath);

	FViews := TObservable.Create;
	FViews.Controller := Self;

	FRefreshed := CreateEventW(nil, True, False, nil);
	{ZbgAlOFF}

	SetCursorWait;
	DoRefresh;
	ResetCursor;
end;

procedure T0Folder.BeforeDestruction;
begin
	FViews.Free;
	FinalizePath(FPath);
	CloseHandle(FRefreshed);
	inherited;
end;

function T0Folder.GetValid: Boolean;
begin
	Result := FValid
end;

procedure T0Folder.Refresh;
var
	OldFiles: TFileArray;
begin
	if FValid then
	begin
		SetCursorWait;

		OldFiles := FFiles; {View ܂QƂĂ\邽߉Ȃ}
		try
			DoRefresh;
		except
			FValid := False;
		end;
		ResetCursor;

		SetEvent(FRefreshed); {XV}[N}

		FViews.Notify; {ʒm}
	end;

	OldFiles := nil; {...}
end;

function T0Folder.GetCount: Integer;
begin
	Result := Length(FFiles)
end;

function T0Folder.GetItems(AIndex: Integer): PFile;
begin
	Result := @FFiles[AIndex];
end;

function T0Folder.GetPath: TPath;
begin
	Result := FPath
end;

procedure T0Folder.ResumeSurveillance(Wait: Boolean);
begin
	{sv}

	if FWatchThread <> nil then
	begin
		ResetEvent(FRefreshed);
		SetEvent(TDirWatchThread(FWatchThread).Surveillance);
		if Wait then
		begin
			Application.Async;
			WaitForSingleObject(FRefreshed, 1000);
			Application.RestoreAsync;
		end
	end
end;

procedure T0Folder.StopSurveillance;
begin
	{sv}

	if FWatchThread <> nil then
		ResetEvent(TDirWatchThread(FWatchThread).Surveillance);
end;

procedure T0Folder.Attach(const AView: IObserver);
begin
	FViews.Attach(AView)
end;

procedure T0Folder.Detach(const AView: IObserver);
begin
	FViews.Detach(AView);
	if not FViews.HasClients then ViewDecreasedAll
end;

function T0Folder.HasClients: Boolean;
begin
	Result := FViews.HasClients
end;

procedure T0Folder.ViewDecreasedAll;
begin
end;

function T0Folder.GetInfoText(AFile: PFile): WideString;
begin
	Result := SReady
end;

{ TDirectory }

constructor TDirectory.Create(const ADirectory: TWideFileName);
var
	P: TPath;
begin
	InitializePath(P);
	P.Directory := ADirectory;
	inherited Create(P)
end;

procedure TDirectory.AfterConstruction;
begin
	inherited;
	BeginWatch;
end;

procedure TDirectory.BeforeDestruction;
begin
	EndWatch;
	inherited;
end;

procedure TDirectory.BeginWatch;
begin
	if not (GetDriveTypeW(PWideChar(ExtractFileDrive(Path.Directory))) in [DRIVE_REMOVABLE, DRIVE_CDROM]) then
		FWatchThread := TDirWatchThread.Create(Self);
end;

procedure TDirectory.EndWatch;
begin
	if FWatchThread <> nil then
	begin
		TDirWatchThread(FWatchThread).Join;
		FWatchThread := nil;
	end
end;

procedure TDirectory.DoRefresh;
var
	FindData: WIN32_FIND_DATAW;
	H: THandle;
	C: Integer;
	Mask: TWideFileName;
	Manager: IFolderManager;
	Attr: DWORD;
begin
	Attr := GetFileAttributesW(Pointer(Path.Directory)); {G[_CAO͂̍sŏo}
	FValid := (Attr <> High(Attr)) and (Attr and faDirectory <> 0) and FileSystemExists(Path.Directory);
	if FValid then
	begin
		Manager := Application as IFolderManager;

		FAtRoot := WinTips.IsRootDir(Path.Directory);

		Mask := IncludeTrailingPathDelimiter(Path.Directory) + '*.*';

		H := FindFirstFileW(Pointer(Mask), FindData);
		if H <> INVALID_HANDLE_VALUE then
		begin
			C := 0;
			SetLength(FFiles, 1024);
			try
				repeat
					if FindData.cFileName[0] = '.' then
          begin
						if FindData.cFileName[1] = #0 then Continue;
						if FindData.cFileName[1] = '.' then
						begin
							if FindData.cFileName[2] = #0 then Continue;
						end;
					end;

					if C >= Length(FFiles) then
						SetLength(FFiles, Length(FFiles) + 128);

					FindDataToFile(FFiles[C], FindData, Manager);
					Inc(C);
				until not FindNextFileW(H, FindData);
			finally
				Windows.FindClose(H);
			end;
			SetLength(FFiles, C);
		end;
	end
end;

function TDirectory.GetAtRoot: Boolean;
begin
	Result := FAtRoot
end;

function TDirectory.GetDisplayText: WideString;
begin
	Result := Path.Directory
end;

function TDirectory.GetParentPath: TPath;
begin
	InitializePath(Result);

	Result.Directory := ExtractFileDir(FPath.Directory);
	if Result.Directory = Path.Directory then Result.Directory := '';

	{UNCbΉ}
	if (Length(Path.Directory) > 2) and
					(Path.Directory[1] = '\') and (Path.Directory[2] = '\') and
					(Length(ExtractFileDrive(Path.Directory)) = Length(Path.Directory)) then
		SHGetSpecialFolderLocation(0, CSIDL_NETWORK, Result.ItemIDList)
	else if Result.Directory = '' then
		SHGetSpecialFolderLocation(0, CSIDL_DRIVES, Result.ItemIDList);
end;

function TDirectory.IsSub(const AParent: TPath): Boolean;
begin
	Result := DirectoryIsSub(Path.Directory, AParent)
end;

procedure TDirectory.Move(const NewPath: WideString);
begin
	EndWatch;
	FPath.Directory := NewPath;
	BeginWatch;
end;

function TDirectory.GetPossibilities: TFolderPossibilities;
begin
	Result := [fpSort, fpPlaceFile, fpHasParent];
end;

function TDirectory.ConcatenatePath(AFile: PFile): TPath;
begin
	InitializePath(Result);
	Result.Directory := ConcatPath(FPath.Directory, AFile.Name);
end;

{ TNameSpace }

constructor TNameSpace.Create(AItemIDList: PItemIDList);
var
	ThePath: TPath;
	Desktop: IShellFolder;
begin
	SHGetDesktopFolder(Desktop);

	InitializePath(ThePath);
	ThePath.ItemIDList := CopyItemIDList(AItemIDList);
	ThePath.Directory := GetDisplayName(Desktop, AItemIDList, SHGDN_FORPARSING);

	FIsRecycleBox := ThePath.Directory = BitbucketPath;
	FIsDrives := ThePath.Directory = DrivesPath;

	inherited Create(ThePath);
end;

procedure TNameSpace.AfterConstruction;
begin
	inherited;

	if FPath.ItemIDList.mkid.cb = 0 then
		FWatchThread := TDirWatchThread.Create(Self);
end;

procedure TNameSpace.BeforeDestruction;
var
	I: Integer;
begin
	if FWatchThread <> nil then
		TDirWatchThread(FWatchThread).Join;

	for I := Low(FFiles) to High(FFiles) do
		CoTaskMemFree(FFiles[I].Data);

	inherited;
end;

function NameSpaceComp(I1, I2: Integer; Data: Pointer): Integer; cdecl;
begin
	Result := SmallInt(TNameSpace(Data).FShellFolder.CompareIDs(0,
					TNameSpace(Data).FFiles[I1].Data,
					TNameSpace(Data).FFiles[I2].Data));
end;

procedure NameSpaceSwap(I1, I2: Integer; Data: Pointer); cdecl;
var
	Temp: TFile;
begin
	Temp := TNameSpace(Data).FFiles[I1];
	TNameSpace(Data).FFiles[I1] := TNameSpace(Data).FFiles[I2];
	TNameSpace(Data).FFiles[I2] := Temp;
end;

procedure TNameSpace.DoRefresh;
var
	Desktop: IShellFolder;

	function GetParentFolder: IShellFolder;
	var
		ParentID: PItemIDList;
	begin
		ParentID := CopyItemIDList(Path.ItemIDList);
		StripLastItemID(ParentID);
		if ParentID^.mkid.cb = 0 then
			Result := Desktop
		else
			Desktop.BindToObject(ParentID, nil, IShellFolder, Result);
		CoTaskMemFree(ParentID);
	end;

const
	EnumOption = SHCONTF_FOLDERS or SHCONTF_NONFOLDERS or SHCONTF_INCLUDEHIDDEN;
var
	Parent: IShellFolder;
	Enum: IEnumIDList;
	SelfID, Item: PItemIDList;
	Got, Dummy, Attr: DWORD;
	ExistedCount: Integer;
	Work: WideString;
	P: PFile;
	Data: TWin32FindDataW;
	Temp64: Int64;
	IsFileSystem: Boolean;
	Manager: IFolderManager;
	Buffer: array[0..1023] of WideChar;
	NetResource: NetResourceW;
begin
	SHGetDesktopFolder(Desktop);

	{Name}

	SelfID := GetLastItemID(Path.ItemIDList);

	Parent := GetParentFolder;

	FDisplayName := ShlTips.GetDisplayName(Parent, SelfID, SHGDN_NORMAL);

	Parent := nil;

	{Bind}

	if Path.ItemIDList^.mkid.cb = 0 then
	begin
		FShellFolder := Desktop;
		FValid := True;
	end
	else
	begin
		FValid := Desktop.BindToObject(Path.ItemIDList, nil, IShellFolder, FShellFolder) = S_OK;
		if not FValid then Exit;
	end;

	Manager := Application as IFolderManager;

	{Enum}

	FValid := FShellFolder.EnumObjects(0, EnumOption, Enum) = NOERROR;
	if not FValid and (Length(FPath.Directory) > 2) and (FPath.Directory[1] = '\') and (FPath.Directory[2] = '\') then
	begin
		NetResource.dwType := RESOURCETYPE_DISK;
		NetResource.lpLocalName := nil;
		NetResource.lpRemoteName := PWideChar(FPath.Directory);
		NetResource.lpProvider := nil;
		if WNetAddConnection3W(GetActiveWindow, NetResource, nil, nil, CONNECT_INTERACTIVE) = NOERROR then
			FValid := FShellFolder.EnumObjects(0, EnumOption, Enum) = NOERROR;
  end;
	if not FValid then Exit;

	ExistedCount := 0;
	SetLength(FFiles, 1024);

	while Enum.Next(1, Item, Got) = S_OK do
	begin
		Inc(ExistedCount);
		if ExistedCount > Length(FFiles) then
			SetLength(FFiles, Length(FFiles) + 128);

		P := @FFiles[ExistedCount - 1];

		P^.Data := Item;

		{Name}

		//OutputDebugStringW(PWideChar(ShlUtils.GetDisplayName(FShellFolder, Item, SHGDN_INFOLDER)));

		Work := ShlTips.GetDisplayName(FShellFolder, Item, SHGDN_FORPARSING);
		if (Length(Work) > 3) and (Work[1] = ':') and (Work[2] = ':') then
			Work := Copy(Work, WideLastDelimiter('\', Work) + 1, MaxInt);

		lstrcpyW(P^.Name, PWideChar(Work));
		P^.NameLength := Length(Work);

		{tHg(C:\WINNT\Fonts̃GCAX)Ȃ񂩂SHGetData...ł͎Ȃ}

		IsFileSystem := SHGetDataFromIDListW(FShellFolder, Item, SHGDFIL_FINDDATA, @Data, SizeOf(Data)) = S_OK;
		if not IsFileSystem and (P^.NameLength > 3) and
						(P^.Name[0] <> ':') and (P^.Name[1] = ':') and (P^.Name[2] = '\') then
		begin
			Data.cFileName[0] := #0;
			Windows.FindClose(Windows.FindFirstFileW(P^.Name, Data));
			IsFileSystem := Data.cFileName[0] <> #0;
		end;

		if IsFileSystem then
		begin
			P^.Kind := fkFile;

			if FIsDrives or FIsRecycleBox then
			begin
				Work := ShlTips.GetDisplayName(FShellFolder, Item, SHGDN_INFOLDER);
				lstrcpyW(P^.DisplayName, PWideChar(Work));
				P^.DisplayNameLength := Length(Work);

				Data.cAlternateFileName[0] := #0;
			end
			else
			begin
				SetString(Work, P^.Name, P^.NameLength);
				Work := ExtractFileName(Work);
				lstrcpyW(P^.DisplayName, PWideChar(Work));
				P^.DisplayNameLength := Length(Work);
			end;

			if Data.cAlternateFileName[0] = #0 then
			begin
				SetString(Work, Buffer, GetShortPathNameW(P^.Name, Buffer, Length(Buffer)));
				Work := ExtractFileName(Work);
				lstrcpyW(Data.cAlternateFileName, PWideChar(Work));
			end;

			lstrcpyW(P^.ShortName, Data.cAlternateFileName);
			P^.ShortNameLength := lstrlenW(Data.cAlternateFileName);

      P^.Number := High(Word);

			Int64Rec(Temp64).Lo := Data.nFileSizeLow;
			Int64Rec(Temp64).Hi := Data.nFileSizeHigh;
			P^.Size := Temp64;
			P^.Attributes := Data.dwFileAttributes;

			P^.CreationTime := Data.ftCreationTime;
			P^.LastWriteTime := Data.ftLastWriteTime;
			P^.LastAccessTime := Data.ftLastAccessTime;
		end
		else if (P^.NameLength = 3) and (P^.Name[1] = ':') and (P^.Name[2] = '\') then
		begin
			P^.Kind := fkDrive;

			Work := ShlTips.GetDisplayName(FShellFolder, Item, SHGDN_INFOLDER);
			lstrcpyW(P^.DisplayName, PWideChar(Work));
			P^.DisplayNameLength := Length(Work);

			P^.DriveType := GetDriveTypeW(P^.Name);

			if P^.DriveType = DRIVE_FIXED then
			begin
				GetVolumeInformationW(P^.Name, P^.VolumeLabel, 13, nil, Dummy, Dummy, P^.FileSystem, 13);
				P^.VolumeLabelLength := lstrlenW(P^.VolumeLabel);
				P^.FileSystemLength := lstrlenW(P^.FileSystem);
				P^.Size := DiskSize(Ord(UpCase(Char(P^.Name[0]))) - Ord('A') + 1);
			end
			else
			begin
				{ǂݎ肪Kvȑ̓n[hfBXNȊOȂ}
				P^.FileSystem[0] := #0;
				P^.FileSystemLength := 0;
				P^.VolumeLabelLength := 0;
				P^.Size := -1;
			end;
		end
		else
		begin
			P^.Kind := fkVirtual;

			Work := ShlTips.GetDisplayName(FShellFolder, Item, SHGDN_NORMAL);
			lstrcpyW(P^.DisplayName, PWideChar(Work));
			P^.DisplayNameLength := Length(Work);
		end;

		{}

		Attr := SFGAO_CAPABILITYMASK or SFGAO_FOLDER or SFGAO_FILESYSTEM;
		FShellFolder.GetAttributesOf(1, Item, Attr);
		P^.Operatable := [];

		if Attr and SFGAO_CANDELETE    <> 0 then Include(P^.Operatable, foDelete);
		if Attr and SFGAO_CANMOVE      <> 0 then Include(P^.Operatable, foMove);
		if Attr and SFGAO_CANRENAME    <> 0 then Include(P^.Operatable, foRename);
		if Attr and SFGAO_CANCOPY      <> 0 then Include(P^.Operatable, foCopy);
		if Attr and SFGAO_CANLINK      <> 0 then Include(P^.Operatable, foLink);
		if Attr and SFGAO_HASPROPSHEET <> 0 then Include(P^.Operatable, foProperty);

		if (Attr and SFGAO_FOLDER <> 0) or (Manager.FindArchiver(P^.Name) <> nil) then
		begin
			Include(P^.Operatable, foEnter);

			if (P^.Name[0] = '\') and (P^.Name[1] = '\') and
				(WideStrScan(P^.Name + 3, WideChar('\')) <> nil) and
				(WideStrScan(P^.Name + 3, WideChar(':')) = nil) then
			begin
				P^.Kind := fkFile;
				P^.Attributes := P^.Attributes or faDirectory
			end;
		end;
	end;

	SetLength(FFiles, ExistedCount);

	QuickSort(Low(FFiles), High(FFiles), NameSpaceComp, NameSpaceSwap, Self);
end;

procedure TNameSpace.Move(const NewPath: WideString);
begin
	raise Exception.Create(SNoAllowed);
end;

function TNameSpace.GetDisplayText: WideString;
begin
	if (Path.ItemIDList.mkid.cb > 0) and (FDisplayName <> Path.Directory) and
    ('\\' + FDisplayName <> FPath.Directory)
  then
		Result := FDisplayName + FPath.Directory
	else
		Result := FDisplayName
end;

function TNameSpace.GetParentPath: TPath;
begin
	Result := CopyPath(FPath);
	if Result.ItemIDList.mkid.cb <> 0 then
		StripLastItemID(Result.ItemIDList);
end;

function TNameSpace.IsSub(const AParent: TPath): Boolean;
begin
	if AParent.ItemIDList = nil then
		Result := False
	else
		Result := CompareMem(Path.ItemIDList, AParent.ItemIDList,
						GetItemIDListSize(AParent.ItemIDList) - SizeOf(Word));
end;

function TNameSpace.AtDesktop: Boolean;
begin
	Result := Path.ItemIDList.mkid.cb = 0
end;

function TNameSpace.AtDrives: Boolean;
begin
	Result := Path.Directory = DrivesPath
end;

function TNameSpace.GetPossibilities: TFolderPossibilities;
begin
	if AtDesktop then
		Result := [fpPlaceFile]
	else
		Result := [fpHasParent]
end;

function TNameSpace.ConcatenatePath(AFile: PFile): TPath;
begin
  InitializePath(Result);

	if AFile^.Kind = fkVirtual then
		Result.ItemIDList := ConcatItemIDLists(Path.ItemIDList, AFile^.Data);

	with AFile^ do SetString(Result.Directory, Name, NameLength);
end;

end.

