unit UMain;

{ EOFile by YT }

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
	Windows, Messages, ActiveX, ShlObj, SysUtils, Types,
	SysUtilsW, WinTips, DYAYaml,
	LObjects, LActions, LMenus, LWindows, LAddDlgs,
	EOFileIntf,
	UPath, UFolders, UView, UActions, UActionCommandLine;

type
	TViewThread = class(T0GUIThread, IEOFileThread, IActionManager, IKeyManager)
	private
		FActionManager: TThreadActionManager;
		FFirstPath: TPath;
		FSetupEvent: THandle;
		function GetActionManager: IActionManager;
		function GetKeyManager: IKeyManager;
		{IEOFileThread}
		function GetActiveView: IView;
		procedure SetActiveView(const Value: IView);
		function Enumerate(Proc: CEachView; Context: Pointer): IView;
	protected
		procedure Initialize; override;
		procedure Finalize; override;
	public
		constructor Create(const AFirstPath: TPath; SetupEvent: THandle = 0);
		property ActionManager: IActionManager read GetActionManager implements IActionManager;
		property KeyManager: IKeyManager read GetKeyManager implements IKeyManager;
		function SelectFolder(const Title: WideString; out Path: TPath): Boolean;
	end;

type
	TApplication = class(T0Application, IEOFileGlobal,
		IActionManager, IFolderManager, IExtensionManager, IYamlSerializer)
	private
		FActionManager: TGlobalActionManager;
		FFolderManager: TFolderManager;
		FExtensions: array of IExtension;
		FExtensionModules: array of HMODULE;
		FSetting: UTF8String;
		FSkinTemplate: TSkinTemplate;
		FInitialFilter: WideString;
		FEditor: WideString;
		function GetYamlReader: IYamlSerializer;
		function GetActionManager: IActionManager;
		function GetFolderManager: IFolderManager;
		{IExtensionManager}
		procedure RegisterExtension(const AExtension: IExtension);
		function Enumerate(Proc: CEachExtension; Context: Pointer): IExtension;
		{IEOFileGlobal}
		procedure StartViewThread(const AFirstPath: TPath);
		procedure StartCommandLineThread(const Title, Dir, Initial: WideString; CaretIndex: Integer; SelectedLength: Integer);
		function GetExePath: WideString;
		function GetInitialFilter: WideString;
		function GetEditor: WideString;
		function IEOFileGlobal.ExecuteCommandLine = CommandLine;
		function CommandLine(const Cmd, Dir: WideString; Wait: Boolean = False): Boolean;
	public
		function ExecuteCommandLine(const CmdLine: PWideChar): HWND; overload;
		procedure Initialize; override;
		procedure BeforeDestruction; override;
		procedure ThreadBegun(AThread: TViewThread);
		procedure ThreadEnded(AThread: TViewThread);
		property InitialFilter: WideString read FInitialFilter;
		property Editor: WideString read FEditor;
		property SkinTemplate: TSkinTemplate read FSkinTemplate;
		property ActionManager: IActionManager read GetActionManager implements IActionManager;
		property FolderManager: IFolderManager read GetFolderManager implements IFolderManager;
		property YamlReader: IYamlSerializer read GetYamlReader implements IYamlSerializer;
	end;

function GetApplication: IInterface;
function GetCurrentThread: IInterface;
function GetInterfaceVersion: Cardinal;

implementation

function GetApplication: IInterface;
begin
	Result := LObjects.Application
end;

function GetCurrentThread: IInterface;
begin
	Result := LObjects.CurrentThread
end;

function GetInterfaceVersion: Cardinal;
begin
	Result := EOFileIntf.InterfaceVersion
end;

{ TViewThread }

constructor TViewThread.Create(const AFirstPath: TPath; SetupEvent: THandle);
begin
	FSetupEvent := SetupEvent;
	FFirstPath := CopyPath(AFirstPath);
	inherited Create;
end;

function TViewThread.Enumerate(Proc: CEachView; Context: Pointer): IView;
begin
	Result := TFilerWindow(MainWindow).EnumerateViews(Proc, Context);
end;

procedure TViewThread.Finalize;
begin
	TApplication(LWindows.Application).ThreadEnded(Self);
	OleUninitialize;
	inherited;
	TApplication(LWindows.Application).ProcessIdle;
end;

function TViewThread.GetActionManager: IActionManager;
begin
	Result := FActionManager
end;

function TViewThread.GetActiveView: IView;
begin
	Result := UView.ActiveView
end;

function TViewThread.GetKeyManager: IKeyManager;
begin
	Assert(MainWindow is TFilerWindow);
	Result := TFilerWindow(MainWindow).KeyManager
end;

procedure TViewThread.Initialize;
	procedure ReadSetting(const Setting: UTF8String);
		var Reader: TYamlReader;
		procedure ReadKey(KeyTable: TKeyAssignment);
			function Strip(var S: UTF8String; const P: UTF8String): Boolean;
			begin
				Result := StrLIComp(PChar(S), Pointer(P), Length(P)) = 0;
				if Result then Delete(S, 1, Length(P));
			end;
		var
			Shift: TShiftKeys;
			Key, I: Byte;
			K: UTF8String;
		begin
			while Reader.Next.Kind <> ytkLeave do begin
				if Reader.Next.Kind = ytkValue then begin
					K := Reader.NextName;
					{C}
					Shift := [];
					Key := 0;
					if Strip(K, 'A-') then Include(Shift, ssAlt);
					if Strip(K, 'C-') then Include(Shift, ssControl);
					if Strip(K, 'S-') then Include(Shift, ssShift);
					{Ӗɑg̗JT̐dl}
					if (Length(K) >= 2) and (K[1] = '_') then Delete(K, 1, 1);
					{񂩂L[R[h}
					if CompareText(K, 'Colon') = 0 then
						Key := $BA
					else if CompareText(K, 'Minus') = 0 then
						Key := $BD
					else if CompareText(K, 'CommercialAt') = 0 then
						Key := $C0
					else if CompareText(K, 'YenSign') = 0 then
						Key := $DC
					else if CompareText(K, 'BackSlash') = 0 then
						Key := $E2
					else if CompareText(K, 'LeftButton') = 0 then
						Key := VK_LBUTTON
					else if CompareText(K, 'MiddleButton') = 0 then
						Key := VK_MBUTTON
					else if CompareText(K, 'RightButton') = 0 then
						Key := VK_RBUTTON
					else begin
						for I := 5 to 250 do begin
							if CompareText(KeyNameText(I), K) = 0 then begin
								Key := I;
								Break;
							end;
						end
					end;
					{ANṼZbg}
					ExceptionClass := EAbort;
					try
						try
							KeyTable.Table[Shift, Key] := ActionManager.Actions[YamlStringExtractEscape(Reader.NextValue)];
						except
							on E: EInvalidContainer do ;
						end;
					finally
						ExceptionClass := Exception;
					end;
				end;
				Reader.AdvanceStructure;
			end;
		end;
		procedure ReadMenu(Menu: TAnyMenu);
		var
			Action: IAction;
			Item: TAnyMenuItem;
			A: WideChar;
			T: WideString;
			OwnerDraw: Boolean;
		begin
			Assert(Application is TApplication);
			OwnerDraw := TApplication(Application).FSkinTemplate.OwnerDraw;
			repeat
				if Reader.Next.Kind = ytkEnterMapping then
				begin
					Reader.Advance;

					A := WideChar(Reader.NextName[1]);
					if A = '+' then A := #0;

					T := YamlStringExtractEscape(Reader.NextValue);
					if (Reader.NextValue <> '') and (Reader.NextValue[1] = '"') then
					begin
						Item := CreateMenuItem(Menu, T);
						Item.Accelerator := A;
					end
					else
					begin
						ExceptionClass := EAbort;
						try
							try
								Action := ActionManager.Actions[T];
							except
								Action := ActionManager.Actions['Dummy'];
								A := #0;
							end;
						finally
							ExceptionClass := Exception;
						end;
						Item := CreateMenuItem(Menu, Action, A);
					end;
					Item.OwnerDraw := OwnerDraw;

					Reader.Advance;
					if (Reader.Next.Kind = ytkEnterSequence) and (Reader.NextName = '*') then
					begin
						Reader.Advance;
						ReadMenu(CreateSubMenu(Item));
						Reader.Advance;
					end;
					while Reader.Next.Kind <> ytkLeave do
						Reader.AdvanceStructure;
				end
				else if Reader.Next.Kind = ytkValue then
				begin
					CreateSeparator(Menu).OwnerDraw := OwnerDraw
				end;
				Reader.AdvanceStructure;
			until Reader.Next.Kind = ytkLeave;
		end;
	begin
		if Setting <> '' then begin
			Reader := TYamlReader.Create('EOFile', Setting);
			try
				if Reader.Next.Kind = ytkEnterMapping then begin
					Reader.Advance;
					while Reader.Next.Kind <> ytkLeave do begin
						if (Reader.Next.Kind = ytkEnterMapping) and (Reader.NextName = 'key') then begin
							Reader.Advance;
							ReadKey(TFilerWindow(MainWindow).KeyAssignment);
							Reader.Advance;
						end else if (Reader.Next.Kind = ytkEnterSequence) and (Reader.NextName = 'menu') then begin
							Reader.Advance;
							ReadMenu(MainWindow.Menu);
							Reader.Advance;
						end else begin
							Reader.AdvanceStructure;
						end;
					end;
				end
			finally
				Reader.Free
			end;
		end
	end;
begin
	inherited;
	SetName('View');
	ReadyRendezvous;
	OleInitialize(nil);
	FActionManager := TThreadActionManager.Create(False);
	FActionManager.Owner := Self;
	TFilerWindow.Create(@FFirstPath).Owner := Self;
	FinalizePath(FFirstPath);

	TApplication(LWindows.Application).ThreadBegun(Self);
	try
		ReadSetting(TApplication(Application).FSetting);
		MainWindow.Menu.Update; {ꂪƈuj[\}
	finally
		if FSetupEvent <> 0 then SetEvent(FSetupEvent);
	end;
end;

function TViewThread.SelectFolder(const Title: WideString; out Path: TPath): Boolean;
begin
	Result := UView.SelectFolder(Title, Path);
end;

procedure TViewThread.SetActiveView(const Value: IView);
begin
	if Value <> nil then begin
		Value.SetFocus;
	end;
end;

{ TApplication }

procedure TApplication.BeforeDestruction;
var
	I: Integer;
begin
	if FFolderManager <> nil then begin
		for I := Low(FExtensions) to High(FExtensions) do begin
			FExtensions[I].Finalize;
		end;
		FExtensions := nil;
		for I := Low(FExtensionModules) to High(FExtensionModules) do begin
			FreeLibrary(FExtensionModules[I]);
		end;
		FActionManager.Free;
		FFolderManager.Free;
	end;
	inherited;
	OleUninitialize;
end;

function TApplication.CommandLine(const Cmd, Dir: WideString; Wait: Boolean): Boolean;
begin
	Result := WinTips.ExecuteCommandLine(Cmd, Dir, SW_NORMAL, Wait)
end;

function TApplication.Enumerate(Proc: CEachExtension; Context: Pointer): IExtension;
var
	I: Integer;
begin
	for I := Low(FExtensions) to High(FExtensions) do begin
		Result := FExtensions[I];
		if Proc(Result, Context) then Exit;
	end;
	Result := nil
end;

function TApplication.ExecuteCommandLine(const CmdLine: PWideChar): HWND;
var
	Options: TWideStringDynArray;
	Path: TPath;
	I: Integer;
	Determined: Boolean;
	Default: WideString;
begin
	Options := ParamStrings(CmdLine);
	Default := '.';
	InitializePath(Path);
	Determined := False;
	I := 1;
	while I <= High(Options) do begin
		if Options[I][1] = '/' then Options[I][1] := '-';
		if WideCompareText(Options[I], '-ROOT') = 0 then begin
			FinalizePath(Path);
			InitializePath(Path);
			SHGetSpecialFolderLocation(0, CSIDL_DESKTOP, Path.ItemIDList);
			Determined := True;
		end else if WideCompareText(Options[I], '-CD') = 0 then begin
			Inc(I);
			if (I <= High(Options)) and not Determined then Default := Options[I];
		end else begin
			FinalizePath(Path);
			Path := ParsePath(Options[I]);
			Determined := True;
		end;
		Inc(I);
	end;
	if not Determined then Path := ParsePath(Default);
	Result := TViewThread.Create(Path).MainWindow.Handle;
	FinalizePath(Path);
end;

function TApplication.GetActionManager: IActionManager;
begin
	Result := FActionManager
end;

function TApplication.GetEditor: WideString;
begin
	Result := FEditor
end;
                   
function TApplication.GetExePath: WideString;
begin
	Result := ExePath
end;

function TApplication.GetFolderManager: IFolderManager;
begin
	Result := FFolderManager
end;

function TApplication.GetInitialFilter: WideString;
begin
	Result := FInitialFilter
end;

function TApplication.GetYamlReader: IYamlSerializer;
begin
	DYAYaml.IYamlSerializer(Result) := TYamlReader.Create('EOFile', FSetting);
end;

procedure TApplication.Initialize;
	procedure ReadExtension(const FileName: WideString);
	var
		Module: HMODULE;
	begin
		Module := SafeLoadLibrary(FileName);
		if Module <> 0 then
		begin
			SetLength(FExtensionModules, Length(FExtensionModules) + 1);
			FExtensionModules[High(FExtensionModules)] := Module
		end
	end;
	procedure ReadSetting(const FileName: WideString);
	var
		Reader: TYamlReader;
		procedure ReadSkin(var Skin: TSkinTemplate);
		begin
			while Reader.Next.Kind <> ytkLeave do begin
				IO(Reader, 'owner-draw', Skin.OwnerDraw);
				IO(Reader, 'flat-scroll-bar', Skin.FlatScrollBar);
				IO(Reader, 'short-file-name', Skin.ShortFileName);
				IO(Reader, 'background-color', Skin.BackgroundColor);
				IO(Reader, 'file-name', Skin.FileName);
				IO(Reader, 'system-file', Skin.SystemFile);
				IO(Reader, 'hidden', Skin.Hidden);
				IO(Reader, 'read-only', Skin.ReadOnly);
				IO(Reader, 'directory', Skin.Directory);
				IO(Reader, 'virtual', Skin.Virtual);
				IO(Reader, 'selected', Skin.Selected);
				IO(Reader, 'line', Skin.Line);
				IO(Reader, 'info-text', Skin.InfoText);
				IO(Reader, 'selected-info-text', Skin.SelectedInfoText);
				IO(Reader, 'background', Skin.Background);
				IO(Reader, 'info', Skin.Info);
				IO(Reader, 'name-space', Skin.NameSpace);
				IO(Reader, 'founds', Skin.Founds);
				IO(Reader, 'archive', Skin.Archive);
				IO(Reader, 'face-name', Skin.FaceName);
				Reader.AdvanceStructure;
			end;
		end;
	var
		H: THandle;
	begin
		H := FileOpen(FileName, fmOpenRead or fmShareDenyWrite);
		if H <> INVALID_HANDLE_VALUE then begin
			SetLength(FSetting, GetFileSize(H));
			FileRead(H, Pointer(FSetting)^, Length(FSetting));
			FileClose(H);
			Reader := TYamlReader.Create('EOFile', FSetting);
			try
				if(Reader.Next.Kind = ytkEnterMapping) then begin
					Reader.Advance;
					while Reader.Next.Kind <> ytkLeave do begin
						if (Reader.Next.Kind = ytkEnterSequence) and (Reader.NextName = 'modules') then begin
							Reader.Advance;
							while Reader.Next.Kind <> ytkLeave do begin
								if Reader.Next.Kind = ytkValue then begin
									ReadExtension(ExePath + YamlStringExtractEscape(Reader.NextValue));
								end;
								Reader.AdvanceStructure;
							end;
							Reader.Advance;
						end
						else if (Reader.Next.Kind = ytkEnterMapping) and (Reader.NextName = 'skin') then begin
							Reader.Advance;
							ReadSkin(FSkinTemplate);
							Reader.Advance;
						end else begin
							IO(Reader, 'initial-filter', FInitialFilter);
							IO(Reader, 'editor', FEditor);
							Reader.AdvanceStructure;
						end;
					end;
				end;
			finally
				Reader.Free;
			end
		end;
	end;
var
	PreWindow, NewWindow: HWND;
	Data: TCopyDataStruct;
	I: Integer;
	CmdLineW: WideString;
begin
	inherited Initialize(nil);
	{notify command-line to the existing instance}
	PreWindow := FindWindowW(FilerWindowClassName, nil);
	if PreWindow <> 0 then
	begin
		CmdLineW := GetCommandLineW + ' -cd "' + GetCurrentDir + '"';
		Data.dwData := CmdLineMagicNumber;
		Data.cbData := (Length(CmdLineW) + 1) * SizeOf(WideChar);
		Data.lpData := PWideChar(CmdLineW);
		NewWindow := SendMessageW(PreWindow, WM_COPYDATA, 0, LPARAM(@Data));
		SetForegroundWindow(NewWindow);
	end
	else
	begin
		try
			OleInitialize(nil);
			{FolderManager}
			FFolderManager := TFolderManager.Create;
			FFolderManager.Owner := Self;
			{ActionMangaer}
			FActionManager := TGlobalActionManager.Create(True);
			FActionManager.Owner := Self;
			{default of template}
			FSkinTemplate.ShortFileName := True;
			FSkinTemplate.BackgroundColor := $ffffff;
			FSkinTemplate.FileName   := $000000;
			FSkinTemplate.SystemFile := $770077;
			FSkinTemplate.Hidden     := $007700;
			FSkinTemplate.ReadOnly   := $770000;
			FSkinTemplate.Directory  := $777700;
			FSkinTemplate.Virtual    := $000077;
			FSkinTemplate.Selected   := $ffffff;
			FSkinTemplate.Line := GetSysColor(COLOR_ACTIVEBORDER);
			FSkinTemplate.InfoText := GetSysColor(COLOR_INFOTEXT);
			FSkinTemplate.SelectedInfoText := GetSysColor(COLOR_INFOBK);
			FSkinTemplate.FaceName := 'lr SVbN';
			{read setting.yaml}
			ReadSetting(ExePath + 'setting.yaml');
			ApplicationSkinTemplate := @FSkinTemplate;
			{initialize extensions}
			for I := Low(FExtensions) to High(FExtensions) do begin
				FExtensions[I].Initialize;
			end;
			{startup}
			ExecuteCommandLine(GetCommandLineW);
			SetCurrentDirNull;
		except
			ShowException(ExceptObject, ExceptAddr);
		end;
	end;
end;

procedure TApplication.RegisterExtension(const AExtension: IExtension);
begin
	SetLength(FExtensions, Length(FExtensions) + 1);
	FExtensions[High(FExtensions)] := AExtension;
end;

procedure TApplication.StartCommandLineThread(const Title, Dir, Initial: WideString; CaretIndex, SelectedLength: Integer);
begin
	TCommandLineThread.Create(Title, Dir, Initial, CaretIndex, SelectedLength)
end;

procedure TApplication.StartViewThread(const AFirstPath: TPath);
begin
	TViewThread.Create(AFirstPath)
end;

procedure TApplication.ThreadBegun(AThread: TViewThread);
var
	I: Integer;
	H: IThreadHandler;
begin
	BeginWrite;
	try
		for I := Low(FExtensions) to High(FExtensions) do begin
			if Supports(FExtensions[I], IThreadHandler, H) then begin
				H.ThreadBegun(AThread)
			end
		end
	finally
		EndWrite;
	end;
end;

procedure TApplication.ThreadEnded(AThread: TViewThread);
var
	I: Integer;
	H: IThreadHandler;
begin
	BeginWrite;
	try
		for I := Low(FExtensions) to High(FExtensions) do begin
			if Supports(FExtensions[I], IThreadHandler, H) then begin
				H.ThreadEnded(AThread)
			end
		end
	finally
		EndWrite;
	end;
end;

initialization

	TApplication.Create;

end.
